from functools import wraps
from logging import getLogger
from os import cpu_count
from queue import Queue
from threading import Thread
from typing import Optional, Tuple, Dict, Callable, Any

from surrealist.connections.connection import Connection
from surrealist.enums import Transport
from surrealist.errors import OperationOnClosedConnectionError
from surrealist.result import SurrealResult
from surrealist.surreal import Surreal
from surrealist.utils import DEFAULT_TIMEOUT

CORES_COUNT = cpu_count()
logger = getLogger("surrealist.connection.pool")


def connected_and_pooled(func):
    """
    Wrapper to check if pool is connected ad delegate work to underlying connections
    :param func: function to wrap
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        # args[0] is a self-argument in methods
        if not args[0].is_connected():
            message = "Your pool is already closed"
            logger.error(message, exc_info=False)
            raise OperationOnClosedConnectionError(message)
        return args[0]._execute(func.__name__, *args[1:], **kwargs)

    return wrapped


class Pool:
    """
    Represents a pool of connections, which is creating a bunch of database connections on start and delegating all
    tasks to the first non-busy connection. So, if there are no more connections in the pool, it tries to create a new
    one if the maximum is not exceeded. If the maximum of connections is reached and no more connections to work with -
    client will be blocked until the first connection finishes the task and appears at the pool.
    """

    def __init__(self, first_connection: Connection, url: str, namespace: Optional[str] = None,
                 database: Optional[str] = None, access: Optional[str] = None, credentials: Tuple[str, str] = None,
                 use_http: bool = False, timeout: int = DEFAULT_TIMEOUT, min_connections: int = CORES_COUNT,
                 max_connections: int = 50):
        self._options = {
            "url": url, "namespace": namespace, "database": database, "access": access, "credentials": credentials,
            "use_http": use_http, "timeout": timeout
        }
        self._timeout = timeout
        self._url = url
        self._min = min_connections if min_connections > 1 else 2
        self._max = max_connections if max_connections <= 50 else 50
        self._main = Queue(maxsize=self._max)
        self._main.put_nowait(first_connection)
        self._counter = 1
        self._connected = True
        self._start()

    def transport(self) -> Transport:
        """
        Returns the transport type of the underlying connections
        :return: Transport enum member
        """
        return Transport.HTTP if self._options["use_http"] else Transport.WEBSOCKET

    @property
    def connections_count(self) -> int:
        """
        Get the current number of connections, all of them can be busy now

        :return: number of connections in the pool
        """
        return self._counter

    def _start(self):
        for _ in range(self._min - 1):
            self._create_new_connection()
        logger.info("Created %s connections", self._min)

    def _create_new_connection(self):
        if self._counter < self._max:
            conn = Surreal(**self._options).connect()
            self._main.put_nowait(conn)
            self._counter += 1

    def close(self):
        """
        Closes the pool. You cannot and should not use a Pool object after that
        """
        self._connected = False
        logger.info("Signal to close pool")
        while not self._main.empty():
            # we need to gently wait for connection to finish the task
            conn = self._main.get(timeout=self._timeout)
            conn.close()
        logger.info("The Pool was closed")

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    def is_connected(self) -> bool:
        """
        Checks the pool is still alive and usable

        :return: True if pool is usable, False otherwise
        """
        return self._connected

    def _execute(self, name, *args, **kwargs) -> SurrealResult:
        """
        Here is the main "magic", this method checks if pool is empty (no more free connections) and if so - creates new
        connection. Then it waits until the first non-busy connection and delegates work to it, calling in method.
        After that - always put connection back to pool

        :param name: name of the connection method to call, for example, "query"
        :param args: args to call
        :param kwargs: keyword args to call
        :return: result of the query
        """
        if self._main.empty():
            Thread(target=self._create_new_connection, daemon=True).start()
        connection = self._main.get()
        try:
            result = getattr(connection, name)(*args, **kwargs)
        finally:
            self._main.put_nowait(connection)
        return result

    @connected_and_pooled
    def query(self, query: str, variables: Optional[Dict] = None) -> SurrealResult:
        """
        This method used for execute a custom SurrealQL query

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#query

        For SurrealQL refer to: https://docs.surrealdb.com/docs/surrealql/overview

        Example:
        websocket_connection.query("SELECT * FROM article;") # gets all records from article table
        websocket_connection.query("SELECT * FROM type::table($tb);", {"tb": "article"}) # gets all records from
        article table using variable tb to specify table

        :param query: any SurrealQL query to execute
        :param variables: a set of variables used by the query
        :return: result of request
        """

    @connected_and_pooled
    def db_info(self) -> SurrealResult:
        """
        Returns info about a current database. You should have permissions for this action.

        Actually converts to QL "INFO FOR DB" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :return: full database information
        """

    @connected_and_pooled
    def db_tables(self) -> SurrealResult:
        """
        Returns all tables names in the current database. You should have permissions for this action.

        Actually call **db_info** and parse tables attribute there.

        :return: list of all tables names
        """

    @connected_and_pooled
    def custom_live(self, custom_query: str, callback: Callable[[Dict], Any]) -> SurrealResult:
        """
        This method can be used to initiate custom live query - a real-time selection from a table with filters and
        other features of Live Query. Works only for websockets.

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Note: all results, DIFF, formats etc. should be specified in the query itself
        """

    @connected_and_pooled
    def kill(self, live_query_id: str) -> SurrealResult:
        """
        This method is used to terminate a running live query by id

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill
        """

    @connected_and_pooled
    def count(self, table_name: str) -> SurrealResult:
        """
        Returns records count for given table. You should have permissions for this action.
        Actually converts to QL "SELECT count() FROM {table_name} GROUP ALL;" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/functions/count

        Note: returns zero if table does not exist, if you need to check table existence use **is_table_exists**

        Note: if you specify table_name with recordID like "person:john" you will get count of fields in record

        :param table_name: name of the table
        :return: result containing count, like SurrealResult(id='', error=None, result=[{'count': 1}], time='123.333µs')
        """



from abc import ABC, abstractmethod
from functools import wraps
from logging import getLogger
from typing import Tuple, Dict, Optional, Union, List, Callable, Any

from surrealist.enums import Transport
from surrealist.errors import OperationOnClosedConnectionError, WrongParameterError
from surrealist.record_id import RecordId
from surrealist.result import SurrealResult
from surrealist.utils import (DEFAULT_TIMEOUT, NS, DB, AC, mask_pass, clean_dates, StrOrRecord,
                              get_table_or_record_id)

logger = getLogger("surrealist.connection")
LINK = "https://github.com/kotolex/surrealist?tab=readme-ov-file#recursion-and-json-in-python"


def connected(func):
    """
    Decorator for methods to make sure the underlying connection is alive (connected to DB)

    :param func: method to decorate
    :raise OperationOnClosedConnectionError: if connection is already closed
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        # args[0] is a self-argument in methods
        if not args[0].is_connected():
            message = "Your connection is already closed"
            logger.error(message, exc_info=False)
            raise OperationOnClosedConnectionError(message)
        return func(*args, **kwargs)

    return wrapped


class Connection(ABC):
    """
    Parent for connection objects, contains all public methods to work with API
    """

    def __init__(self, db_params: Optional[Dict] = None, credentials: Tuple[str, str] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        """
        Init any connection to use
        :param db_params: optional parameter, if it is not None, should be like {"NS": "test", "DB": "test"}
        :param credentials: optional pair of user and pass for auth, like ("root", "root")
        :param timeout: timeout in seconds to wait connection results and responses
        """
        self._db_params = db_params
        self._credentials = credentials
        self._connected = False
        self._timeout = timeout
        self._token = None

    def close(self):
        """
        Closes the connection. You can not and should not use a connection object after that
        """
        logger.info("The connection was closed")
        self._connected = False

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    def is_connected(self) -> bool:
        """
        Checks the connection is still alive and usable

        :return: True if connection is usable, False otherwise
        """
        return self._connected

    @connected
    def count(self, table_name: str) -> SurrealResult:
        """
        Returns records count for given table. You should have permissions for this action.
        Actually converts to QL "SELECT count() FROM {table_name} GROUP ALL;" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/functions/count

        Note: returns zero if table does not exist, if you need to check table existence use **is_table_exists**

        Note: if you specify table_name with recordID like "person:john" you will get count of fields in record

        :param table_name: name of the table
        :return: result containing count, like SurrealResult(id='', error=None, result=[{'count': 1}], time='123.333µs')
        """
        logger.info("Query-Operation: COUNT. Table: %s", table_name)
        result = self.query(f"SELECT count() FROM {table_name} GROUP ALL;")
        if not result.is_error():
            result.result = self._get_count(result.result)
        return result

    @connected
    def table_info(self, table_name: str, structured: bool = False) -> SurrealResult:
        """
        Returns info about specified table. You should have permissions for this action.

        Actually converts to QL "INFO FOR TABLE table_name" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :param table_name: name of the table
        :param structured: if True returns data in structured view (use STRUCTURE statement). Note: experimental!
        :return: full table information
        """
        return self._info(f"TABLE {table_name}", structured)

    @connected
    def db_info(self, structured: bool = False) -> SurrealResult:
        """
        Returns info about a current database. You should have permissions for this action.

        Actually converts to QL "INFO FOR DB" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :param structured: if True, return data in structured view (use STRUCTURE statement). Note: experimental!
        :return: full database information
        """
        return self._info("DB", structured)

    @connected
    def ns_info(self, structured: bool = False) -> SurrealResult:
        """
        Returns info about current namespace. You should have permissions for this action.

        Actually converts to QL "INFO FOR NS" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :param structured: if True, return data in structured view (use STRUCTURE statement). Note: experimental!
        :return: full namespace information
        """
        return self._info("NS", structured)

    @connected
    def root_info(self, structured: bool = False) -> SurrealResult:
        """
        Returns info about root. You should have permissions for this action.

        Actually converts to QL "INFO FOR ROOT" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :param structured: if True, return data in structured view (use STRUCTURE statement). Note: experimental!
        :return: information about root
        """
        return self._info("ROOT", structured)

    def _info(self, type_: str, structured: bool = False) -> SurrealResult:
        if structured:
            type_ = f"{type_} STRUCTURE"
        logger.info("Query-Operation: %s_INFO", type_)
        return self.query(f"INFO FOR {type_};")

    @connected
    def session_info(self) -> SurrealResult:
        """
        Returns info about the current session. You should have permissions for this action.

        Actually converts to QL query to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/functions/session

        :return: full session information
        """
        query = 'RETURN {"db" : session::db(), "session_id" : session::id(), "ip" : session::ip(), ' \
                '"ns" : session::ns(), "http_origin" : session::origin(), "access" : session::ac()};'
        logger.info("Query-Operation: SESSION_INFO")
        return self.query(query)

    @connected
    def info(self) -> SurrealResult:
        """
        This method returns the record of an authenticated record user.
        The result property of the response is likely different depending on your schema and the authenticated user

        :return: result of the query
        """
        logger.info("Query-Operation: INFO")
        data = {"method": "info"}
        result = self._use_rpc(data)
        return result

    @connected
    def db_tables(self) -> SurrealResult:
        """
        Returns all tables names in the current database. You should have permissions for this action.

        Actually call **db_info** and parse tables attribute there.

        :return: list of all tables names
        """
        logger.info("Query-Operation: DB_TABLES")
        res: SurrealResult = self.db_info()
        res.result = list(res.result["tables"].keys())
        return res

    @connected
    def is_table_exists(self, table_name: str) -> bool:
        """
        Returns True if table with given name exists in a current database. You should have permissions for this action.

        :param table_name: name of the table, we do not expect record_id here
        :return: True if table exists, False otherwise
        """
        return table_name in self.db_tables().result

    @connected
    def remove_table(self, table_name: str, if_exists: bool = True) -> SurrealResult:
        """
        Fully removes table, even if it contains some records, analog of SQL "DROP table". You should have permissions
        for this action.This method cannot remove any other resource if you need to remove db, ns or access -
        use **query**

        If if_exists parameter is False and the table does not exist - error will be returned at a result.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/remove

        Note: only name of the table allowed here, do not use record_id

        :param table_name: the name of the table,
        :param if_exists: if True, will use IF EXISTS statement for the query
        :return: result of the query
        """
        logger.info("Query-Operation: REMOVE. Table name %s", table_name)
        add = "" if not if_exists else "IF EXISTS"
        return self.query(f"REMOVE TABLE {add} {table_name};")

    @connected
    def show_changes(self, table_name: str, since: str, limit: int = 10) -> SurrealResult:
        """
        This method can use Changes Feed feature of SurrealDB. You need a predefined CHANGEFEED table and database for
        using this action.

        Refer to: https://surrealdb.com/blog/unlocking-streaming-data-magic-with-surrealdb-live-queries-and-change-feeds

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#change-feeds

        :param table_name: name of the table, no record_id expected here
        :param since: str representation of ISO date-time, for example d"2024-02-06T10:48:08.700483Z"
        :param limit: amount of changes to get
        :return: result of the query
        """
        query = f'SHOW CHANGES FOR TABLE {table_name} SINCE {since} LIMIT {limit};'
        return self.query(query)

    @abstractmethod
    def _use_rpc(self, data) -> SurrealResult:
        """
        Actual use of RPC protocol for a current connection type
        """

    def _signin(self, user: str, password: str, namespace: Optional[str] = None, database: Optional[str] = None,
                access: Optional[str] = None) -> SurrealResult:
        """
        This method allows you to sign in a root, namespace, database or scope user against SurrealDB

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#signin

        Example:
        connection.signin('root', 'root') # sign in as root user
        connection.signin('user_db', 'user_db', namespace='test', database='test') # sign in as db user

        :param user: name of the user
        :param password: password for auth
        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :param access: name of the access method to use
        :return: result of request
        """
        params = {"user": user, "pass": password}
        if user is None or password is None:
            params = {}
        if namespace is not None:
            params[NS] = namespace
        if database is not None:
            params[DB] = database
        if access is not None:
            params[AC] = access
        data = {"method": "signin", "params": [params]}
        logger.info("Operation: SIGNIN. Data: %s", mask_pass(str(params)))
        return self._use_rpc(data)

    @abstractmethod
    def use(self, namespace: str, database: Optional[str] = None) -> SurrealResult:
        """
        This method specifies the namespace and optionally database for the current connection

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/use
        """

    @abstractmethod
    def transport(self) -> Transport:
        """
        This method returns the transport type for the current connection

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#transports
        """

    @connected
    def let(self, name: str, value: Any) -> SurrealResult:
        """
        This method sets and stores a value which can then be used in a subsequent query.
        Http-transport cannot use the let method

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#let

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/let

        :param name: name for the variable (without $ sign!)
        :param value: value for the variable
        :return: result of request
        """
        data = {"method": "let", "params": [name, value]}
        logger.info("Operation: LET. Name: %s, Value: %s", name, value)
        return self._use_rpc(data)

    @connected
    def unset(self, name: str) -> SurrealResult:
        """
        This method unsets value, which was previously stored.
        Http-transport cannot use the unset method

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#unset

        :param name: name for the variable (without $ sign!)
        :return: result of request
        """
        data = {"method": "unset", "params": [name]}
        logger.info("Operation: UNSET. Variable name: %s", name)
        return self._use_rpc(data)

    @abstractmethod
    def live(self, table_name: str, callback: Callable[[Dict], Any], return_diff: bool = False) -> SurrealResult:
        """
        This method can be used to initiate live query - a real-time selection from a table. Works only for websockets.

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        About DIFF refer to: https://jsonpatch.com

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query
        """

    @abstractmethod
    def custom_live(self, custom_query: str, callback: Callable[[Dict], Any]) -> SurrealResult:
        """
        This method can be used to initiate custom live query - a real-time selection from a table with filters and
        other features of Live Query. Works only for websockets.

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Note: all results, DIFF, formats etc. should be specified in the query itself
        """

    @abstractmethod
    def kill(self, live_query_id: str) -> SurrealResult:
        """
        This method is used to terminate a running live query by id

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill
        """

    @connected
    def graphql(self, query: Dict, pretty: Optional[bool] = False) -> SurrealResult:
        """
        This method allows you to execute GraphQL queries against the database.
        The query parameter is a dictionary with the following fields:
        - query (required): The GraphQL query string.
        - variables or vars (optional): An object containing variables for the query.
        - operationName or operation (optional): The name of the operation to execute.

        Refer to: https://surrealdb.com/docs/surrealdb/integration/rpc#graphql

        Refer to: https://surrealdb.com/docs/surrealdb/querying/graphql

        Important Note: GraphQL validates all schemas for all tables in the database, so if there are some errors,
        you get an error back, even if the problem is not with your data

        Examples:
        connection.graphql({"query": "{ author { id name } }"}, pretty=True)

        :param query: dictionary with all parameters
        :param pretty: optional boolean parameter, indicating whether the output should be pretty-printed.
        :return: result of request
        :raise WrongParameterError: if query is not valid dictionary
        """
        allowed_fields = ("query", "variables", "vars", "operationName", "operation")
        if "query" not in query or any(field not in allowed_fields for field in query.keys()):
            raise WrongParameterError("Query parameter should be a dictionary with 3 fields\n"
                                      "Please see https://surrealdb.com/docs/surrealdb/integration/rpc#graphql")
        data = {"method": "graphql", "params": [query, {"pretty": pretty}]}
        logger.info("Operation: GRAPHQL. Query: %s, pretty: %s", query, pretty)
        result = self._use_rpc(data)
        return result

    @connected
    def run(self, func_name: str, version: Optional[str] = None, args: Optional[List] = None) -> SurrealResult:
        """
        This method allows you to execute built-in functions, custom functions, or machine learning models with
        optional arguments

        Refer to: https://surrealdb.com/docs/surrealdb/integration/rpc#run

        Examples:
        connection.run("time::now")

        :param func_name: The name of the function or model to execute. Prefix with fn:: for custom functions or
        ml:: for machine learning models.
        :param version: optional parameter, the version of the function or model to execute. When using a machine
        learning model (prefixed with ml::), the version parameter is required.
        :param args: The list of arguments to pass to the function or model.
        :return: result of request
        """
        data = {"method": "run", "params": [func_name]}
        if version is not None:
            data["params"].append(version)
        if args is not None:
            if len(data["params"]) == 1:
                data["params"].append(None)
            data["params"].append(args)
        logger.info("Operation: RUN. Function: %s, version: %s, args: %s", func_name, version, args)
        result = self._use_rpc(data)
        return result

    @connected
    def version(self) -> SurrealResult:
        """
        This method returns version information about the database/server

        Refer to: https://surrealdb.com/docs/surrealdb/integration/rpc#version

        Examples:
        connection.version() # get version information

        :return: result of request
        """
        data = {"method": "version"}
        logger.info("Operation: VERSION")
        result = self._use_rpc(data)
        return result

    @connected
    def select(self, table_name: str, record_id: Optional[StrOrRecord] = None) -> SurrealResult:
        """
        This method selects either all records in a table or a single record

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#select

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/select

        Examples:
        connection.select("article") # select all records in article table
        connection.select("article:first") # select one article with id 'first'
        connection.select("article", "first") # select one article with id 'first', analog of previous
        connection.select("article", RecordId("first", "article")) # analog of previous

        Notice: do not specify id twice: in table name and in record_id, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to select
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = get_table_or_record_id(table_name, record_id)
        data = {"method": "select", "params": [table_name]}
        logger.info("Operation: SELECT. Table: %s", table_name)
        result = self._use_rpc(data)
        if not isinstance(result.result, List) and not result.is_error():
            result.result = [result.result] if result.result else []
        return result

    @connected
    def create(self, table_name: str, data: Dict, record_id: Optional[StrOrRecord] = None) -> SurrealResult:
        """
        This method creates a record either with a random or specified record_id. If no id specified in record_id or
        in data arguments, then id will be generated by SurrealDB

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#create

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/create

        Examples:
        connection.create("person", {"name": "John Doe"}) # create one record in person table with random id
        connection.create("person", {"id":"my_id", "name": "John Doe"}) # create one record in person table
        with specified id
        connection.create("person", {"name": "John Doe"}, "my_id") # create one record in person table with specified id
        connection.create("person:my_id", {"name": "John Doe"}) # create one record in person table with specified id

        Notice: do not specify id twice, for example, in table name and in data, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to create
        :param data: dict with data to create
        :param record_id: optional parameter, it can be string or record_id object
        :return: result of request
        """
        if record_id is not None:
            if isinstance(record_id, str):
                record_id = RecordId(record_id, table=table_name)
            data["id"] = record_id.to_valid_string()
        _data = {"method": "create", "params": [table_name, data]}
        logger.info("Operation: CREATE. Table: %s, data: %s", table_name, data)
        result = self._use_rpc(_data)
        if isinstance(result.result, List) and len(result.result) == 1:
            result.result = result.result[0]
        return result

    @connected
    def update(self, table_name: str, data: Dict, record_id: Optional[StrOrRecord] = None) -> SurrealResult:
        """
        This method can be used to update or modify records in the database. So all old fields will be deleted, and new
        will be added, if you wand just to add field to record, keeping old ones -use **merge** method instead.
        If a record with specified id does not exist, it will NOT be created, use **upsert** for that.

        Note: if you want to create/replace one record, you should specify recordID in table_name or in record_id, but
        not in data parameters.

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#update

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update

        Example:
        connection.update("person:my_id", {"name": "Alex Doe"}) # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted
        connection.update("person", {"name": "Alex Doe"}, "my_id") # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted

        Notice: do not specify id twice, for example, in table name and in record_id, it will cause error

        :param table_name: table name or table name with record_id to update
        :param data: dict with data to create
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = get_table_or_record_id(table_name, record_id)
        _data = {"method": "update", "params": [table_name, data]}
        logger.info("Operation: UPDATE. Table: %s, data: %s", table_name, data)
        return self._use_rpc(_data)

    @connected
    def upsert(self, table_name: str, data: Dict, record_id: Optional[StrOrRecord] = None) -> SurrealResult:
        """
        This method can be used to create or update records in the database. So all old fields will be deleted, and new
        will be added, if you wand just to add field to record, keeping old ones -use **merge** method instead.
        If a record with specified id does not exist, it will be created, if it exists - all fields will be replaced

        Note: if you want to create/replace one record, you should specify recordID in table_name or in record_id, but
        not in data parameters.

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#upsert

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/upsert

        Example:
        connection.upsert("person:my_id", {"name": "Alex Doe"}) # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted

        Notice: do not specify id twice, for example, in table name and in record_id, it will cause error

        :param table_name: table name or table name with record_id to upsert
        :param data: dict with data to create
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = get_table_or_record_id(table_name, record_id)
        _data = {"method": "upsert", "params": [table_name, data]}
        logger.info("Operation: UPSERT. Table: %s, data: %s", table_name, data)
        return self._use_rpc(_data)

    @connected
    def insert(self, table_name: str, data: Union[List, Dict]) -> SurrealResult:
        """
        This method inserts one or more records. If you specify recordID in data and record with that id already
        exists - no inserts or updates will happen and the content of the existing record will be returned. If you need
        to change existing record, please consider **update** or **merge**

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#insert

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/insert

        Examples:
        connection.insert("person", {"name": "John Doe"}) # inserts one record with random id
        connection.insert("person", [{"name": "John Doe"}, {"name", "Jane Doe"}]) # inserts two records with random ids

        Note: do not use record id in table_name parameter (table:recordID) - it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to insert
        :param data: dict or list(many records) with data to create
        :return: result of request
        """
        _data = {"method": "insert", "params": [table_name, data]}
        logger.info("Operation: INSERT. Table: %s, data: %s", table_name, data)
        return self._use_rpc(_data)

    @connected
    def insert_relation(self, table_name: Optional[str], data: [Dict]) -> SurrealResult:
        """
        This method inserts a new relation record into the database. You can specify the relation table to insert into
        and provide the data for the new relation.

        Refer to: https://surrealdb.com/docs/surrealdb/integration/rpc#insert_relation

        Examples:
        connection.insert_relation("likes",{
            "in": "user:alice",
            "out": "post:123",
            "since": "2024-09-15T12:34:56Z"
        })

        :param table_name: The name of the relation table to insert into. If None, the table is determined from the id
        field in the data.
        :param data: dict containing the data for the new relation record, including in, out, and any additional fields
        :return: result of request
        """
        data = {"method": "insert_relation", "params": [table_name, data]}
        logger.info("Operation: INSERT-RELATION. Table: %s, data: %s", table_name, data)
        result = self._use_rpc(data)
        return result

    @connected
    def merge(self, table_name: str, data: Dict, record_id: Optional[StrOrRecord] = None) -> SurrealResult:
        """
        This method merges specified data into either all records in a table or a single record. Old data in records
        will not be deleted, if you want to replace old data with new - use **update** method.If
        record with specified id does not exist, it will be created.

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#merge

        Examples:
        connection.merge("person",{"active": True}) # "active" will be added to all records in person table
        connection.merge("person:my_id", {"active": True}) # "active" will be added to one record in person
        table with specified id

        :param table_name: table name or table name with record_id to merge
        :param data: dict with data to add
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = get_table_or_record_id(table_name, record_id)
        _data = {"method": "merge", "params": [table_name, data]}
        logger.info("Operation: MERGE. Table: %s, data: %s", table_name, data)
        return self._use_rpc(_data)

    @connected
    def delete(self, table_name: str, record_id: Optional[StrOrRecord] = None) -> SurrealResult:
        """
        This method deletes all records in a table or a single record, be careful and don't forget to specify id if you
        do not want to delete all records. This method does not remove table itself, only records in it.As a result of
        this method you will get all deleted records or None if no such record or table

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#delete

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/delete

        Examples:
        connection.delete("person:my_id") # deletes one record in person table
        connection.delete("person", "my_id") # deletes one record in person table
        connection.delete("person") # deletes all records in person table

        :param table_name: table name or table name with record_id to delete
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = get_table_or_record_id(table_name, record_id)
        _data = {"method": "delete", "params": [table_name]}
        logger.info("Operation: DELETE. Table: %s", table_name)
        return self._use_rpc(_data)

    @connected
    def patch(self, table_name: str, data: Union[Dict, List], record_id: Optional[StrOrRecord] = None,
              return_diff: bool = False) -> SurrealResult:
        """
        This method changes specified data in one ar all records. If given table does not exist, new table and record
        will not be created if table exists but no such record_id - new record will be created, if no record id-all
        records will be transformed

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#patch

        About allowed data format and DIFF refer to: https://jsonpatch.com

        Examples:
        connection.patch("person", [{"op": "replace", "path": "/active", "value": False}]) # replaces active
        field for all records in person table to False
        connection.patch("person:my_id", [{"op": "replace", "path": "/active", "value": False}]) # replaces
        active field for one record with specified id to False

        Notice: do not specify id twice, for example, in table name and in data, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to patch
        :param data: list with json-patch data
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :param return_diff: True if you want to get only DIFF info, False for a standard results
        :return: result of request
        """
        table_name = get_table_or_record_id(table_name, record_id)
        params = [table_name, data]
        if return_diff:
            params.append(return_diff)
        _data = {"method": "patch", "params": params}
        logger.info("Operation: PATCH. Table: %s, data: %s, use DIFF: %s", table_name, data, return_diff)
        return self._use_rpc(_data)

    @connected
    def query(self, query: str, variables: Optional[Dict] = None) -> SurrealResult:
        """
        This method used for execute a custom SurrealQL query

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#query

        For SurrealQL refer to: https://docs.surrealdb.com/docs/surrealql/overview

        Example:
        connection.query("SELECT * FROM article;") # gets all records from article table
        connection.query("SELECT * FROM type::table($tb);", {"tb": "article"}) # gets all records from
        article table using variable tb to specify table

        :param query: any SurrealQL query to execute
        :param variables: a set of variables used by the query
        :return: result of request
        """
        query = clean_dates(query)
        params = [query]
        if variables is not None:
            params.append(variables)
        data = {"method": "query", "params": params}
        logger.info("Operation: QUERY. Query: %s, variables: %s", query, variables)
        result = self._use_rpc(data)
        result.query = params[0] if len(params) == 1 else params
        return result

    @connected
    def relate(self, relate_to: str, relation_table: str, relate_from: str,
               data: Optional[Dict] = None) -> SurrealResult:
        """
        This method relates two records with a specified relation

        Refer to: https://surrealdb.com/docs/surrealdb/integration/rpc#relate

        Examples:
        connection.relate("person:john", "knows", "person:jane")

        :param relate_to: The record to relate to
        :param relation_table: name of the relation table
        :param relate_from: The record to relate from
        :param data: dict containing the data for the new record
        :return: result of request
        """
        full_data = {"method": "relate", "params": [relate_to, relation_table, relate_from]}
        if data is not None:
            full_data["params"].append(data)
        logger.info("Operation: RELATE. Relate_to: %s, relation_table: %s, relate_from: %s, data: %s", relate_to,
                    relation_table, relate_from, data)
        result = self._use_rpc(full_data)
        return result

    @abstractmethod
    def import_data(self, path) -> SurrealResult:
        """
        This method imports a SurrealQL script file into a local or remote SurrealDB database server.

        Note: websocket connection cannot use this method

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#import
        https://docs.surrealdb.com/docs/cli/import
        """

    @abstractmethod
    def export(self) -> str:
        """
        This method exports all data for a specific namespace and database.

        Note: websocket connection cannot use this method

        Refer to: https://docs.surrealdb.com/docs/integration/http#export

        Refer to: https://docs.surrealdb.com/docs/cli/export/
        """

    @abstractmethod
    def ml_import(self, path) -> SurrealResult:
        """
        This method imports a SurrealQL ML file into a local or remote SurrealDB database server.

        Note: websocket connection cannot use this method

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#ml-import
        https://docs.surrealdb.com/docs/cli/ml/import
        """

    @abstractmethod
    def ml_export(self, name: str, version: str) -> str:
        """
        This method exports a SurrealML machine learning model from a specific namespace and database.
        As machine learning files can be large, the endpoint outputs a chunked HTTP response.

        Note: websocket connection cannot use this method

        Refer to: https://docs.surrealdb.com/docs/integration/http#ml-export

        Refer to: https://docs.surrealdb.com/docs/cli/ml/export
        """

    def _get_count(self, res) -> int:
        if not res:
            return 0
        if 'count' in res:
            return res['count']
        if 'result' in res:
            return self._get_count(res["result"])
        return self._get_count(res[0])



from typing import List

from surrealist.connections import Connection
from surrealist.utils import OK
from .permissions import CanUsePermissions, CanUseComment
from .statement import Statement


class Alter(Statement, CanUsePermissions, CanUseComment):
    """
    Represents ALTER statement, it should be able to use any statements from documentation

    Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/alter

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_alter_examples.py

    ALTER [
    | TABLE [ IF NOT EXISTS ] @name
    [ DROP ]
        [ SCHEMAFULL | SCHEMALESS ]
        [ PERMISSIONS [ NONE | FULL
            | FOR select @expression
            | FOR create @expression
            | FOR update @expression
            | FOR delete @expression
        ] ]
    [ COMMENT @string ]
    ]
    """

    def __init__(self, connection: Connection, name: str):
        super().__init__(connection)
        self._type = "TABLE"
        self._name = name
        self._if_not_exists = None
        self._drop = None
        self._schema_full = None
        self._as_select = None

    def validate(self) -> List[str]:
        return [OK]

    def if_not_exists(self) -> "Alter":
        """
        Represents IF NOT EXISTS statement
        """
        self._if_not_exists = True
        return self

    def drop(self) -> "Alter":
        """
        Represents DROP statement
        """
        self._drop = True
        return self

    def schemafull(self) -> "Alter":
        """
        Represents SCHEMAFULL statement
        """
        self._schema_full = True
        return self

    def schemaless(self) -> "Alter":
        """
        Represents SCHEMALESS statement
        """
        self._schema_full = False
        return self

    def _clean_str(self):
        exists = "" if not self._if_not_exists else " IF NOT EXISTS"
        drop = "" if not self._drop else " DROP"
        schema = ""
        if self._schema_full is True:
            schema = " SCHEMAFULL"
        elif self._schema_full is False:
            schema = " SCHEMALESS"
        return f"ALTER {self._type}{exists} {self._name}{drop}{schema}"



import urllib.parse
from logging import getLogger
from pathlib import Path
from typing import Optional, Tuple, Dict, Union, Callable, Any

from surrealist.clients.ws_client import WebSocketClient
from surrealist.connections.connection import Connection, connected
from surrealist.enums import Transport
from surrealist.errors import (SurrealConnectionError, WebSocketConnectionError, WebSocketConnectionClosedError,
                               CompatibilityError)
from surrealist.result import SurrealResult
from surrealist.utils import DEFAULT_TIMEOUT, NS, DB, AC

logger = getLogger("surrealist.connections.websocket")


class WebSocketConnection(Connection):
    """
    Represents websocket transport and abilities to work with SurrealDb. It is a recommended connection.

    Refer to surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#transports

    Refer to: https://docs.surrealdb.com/docs/integration/websocket

    Each object creates only one websocket connection and can be used in the context manager to close properly.
    You cannot and should not try to use this object after closing connection. Just create a new connection.

    On creating, this object tries to create connection with specified parameters and will raise exception on fail.
    If namespace and database specified - USE method are called automatically
    If credentials specified - SIGNIN are called automatically
    """

    def __init__(self, url: str, db_params: Optional[Dict] = None, credentials: Optional[Tuple[str, str]] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        super().__init__(db_params, credentials, timeout)
        self._url = url
        base_url = urllib.parse.urlparse(url.lower())
        self._db_params = {}
        if db_params:
            if NS in db_params:
                self._db_params[NS] = db_params[NS]
            if DB in db_params:
                self._db_params[DB] = db_params[DB]
            if AC in db_params:
                self._db_params[DB] = db_params[AC]
        if base_url.scheme in ("ws", "wss"):
            self._base_url = url
        if base_url.scheme in ("http", "https"):
            self._base_url = f"{base_url.scheme.replace('http', 'ws')}://{base_url.netloc}/rpc"
        try:
            self._client = WebSocketClient(self._base_url, timeout)
        except TimeoutError:
            logger.error("Cant connect to %s in %s seconds", self._base_url, self._timeout)
            raise SurrealConnectionError(f"Cant connect to {self._base_url} in {timeout} seconds.\n"
                                         f"Is your SurrealDB started and work on that url? "
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")
        except WebSocketConnectionClosedError:
            logger.error("Cant connect to %s, connection refused", self._base_url)
            raise SurrealConnectionError(f"Cant connect to {self._base_url}, connection refused.\n"
                                         f"Is your SurrealDB started and work on that url? "
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")
        self._use_or_sign_on_params(credentials)
        self._connected = True
        masked_creds = None if not credentials else (credentials[0], "******")
        logger.info("Connected to %s, params: %s, credentials: %s, timeout: %s", self._base_url, db_params,
                    masked_creds, timeout)

    def _use_rpc(self, data) -> SurrealResult:
        return self._run(data)

    def _use_or_sign_on_params(self, credentials):
        if self._db_params:
            ns = self._db_params.get(NS)
            db = self._db_params.get(DB)
            ac = self._db_params.get(AC)
            if credentials:
                self._user, self._pass = self._credentials
                signin_result = self._signin(self._user, self._pass, ns, db, ac)
                if signin_result.is_error():
                    logger.error("Error on connecting to %s. Info %s", self._base_url, signin_result)
                    raise WebSocketConnectionError(f"Error on connecting to {self._base_url}.\n"
                                                   f"Info: {signin_result.result}")
                self._token = signin_result.result
            else:
                use_result = self.use(ns, db)
                if use_result.is_error():
                    logger.error("Error on use %s. Info %s", self._db_params, use_result)
                    raise WebSocketConnectionError(f"Error on use '{self._db_params}'.\nInfo: {use_result.result}")
        else:
            if credentials:
                self._user, self._pass = self._credentials
                signin_result = self._signin(self._user, self._pass)
                if signin_result.is_error():
                    logger.error("Error on connecting to %s. Info %s", self._base_url, signin_result)
                    raise WebSocketConnectionError(f"Error on connecting to '{self._base_url}'.\nInfo: {signin_result}")
                self._token = signin_result.result

    def transport(self) -> Transport:
        """
        Returns the transport type for websocket connection
        """
        return Transport.WEBSOCKET

    @connected
    def use(self, namespace: str, database: Optional[str] = None) -> SurrealResult:
        """
        This method specifies the namespace and database for the current connection.
        For websocket connection, you should specify both.

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#use

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/use

        Example:
        websocket_connection.use("test", "test") # use test namespace and test database for this connection

        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :return: result of request
        """
        if not database:
            # For some reason, websocket connection cannot work with only namespace
            msg = "Both namespace and database are required"
            logger.error(msg)
            raise CompatibilityError(msg)
        params = [namespace, database]
        data = {"method": "use", "params": params}
        logger.info("Operation: USE. Namespace: %s, database %s", namespace, database or "None")
        result = self._run(data)
        if not result.is_error():
            # if USE was OK we need to store new data (ns and db)
            self._db_params = {"NS": namespace, "DB": database}
        return result

    @connected
    def live(self, table_name: str, callback: Callable[[Dict], Any], return_diff: bool = False) -> SurrealResult:
        """
        This method can be used to initiate live query - a real-time selection from a table

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#live

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        About DIFF refer to: https://jsonpatch.com

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Example:
        websocket_connection.live("article", callback=lambda a_dict: print(a_dict)) # creates a live query to check any
        events on article table and just printing all incoming results

        :param table_name: name of the table to observe
        :param callback: a function to call on any incoming event. It should take one argument - a dict
        :param return_diff: True if you want to get only DIFF info on table events, False for a standard results
        :return: result of request with the live_id in 'result' field
        """
        params = [table_name]
        if return_diff:
            params.append(True)
        data = {"method": "live", "params": params}
        logger.info("Operation: LIVE. Data: %s", params)
        return self._run(data, callback)

    @connected
    def custom_live(self, custom_query: str, callback: Callable[[Dict], Any]) -> SurrealResult:
        """
        This method can be used to initiate custom live query - a real-time selection from a table with filters and
        other features of Live Query

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Note: all results, DIFF, formats etc. should be specified in the query itself

        Example:
        ws.custom_live("LIVE SELECT * FROM person WHERE age > 18;", callback=lambda a_dict: print(a_dict)) # creates
        a live query to check any events on person table with records, where age field is bigger than 18, and just
        printing all incoming results

        :param custom_query: full LIVE SELECT query text
        :param callback: a function to call on any incoming event. It should take one argument - a dict
        :return: result of request with the live_id in 'result' field
        """
        data = {"method": "query", "params": [custom_query], "additional": "live"}
        logger.info("Operation: CUSTOM LIVE. Query: %s", custom_query)
        result = self._run(data, callback)
        result.query = custom_query
        return result

    @connected
    def kill(self, live_query_id: str) -> SurrealResult:
        """
        This method is used to terminate a running live query by id

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#kill

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill

        :param live_query_id: id for the query to kill
        :return: result of request
        """
        data = {"method": "kill", "params": [live_query_id]}
        logger.info("Operation: KILL. Live_id: %s", live_query_id)
        return self._run(data)

    def export(self):
        """
        Websocket transport cannot use export operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/export

        :raise CompatibilityError: on any use
        """
        message = "Export is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/export"
        raise CompatibilityError(full_message)

    def ml_export(self, _name: str, _version: str):
        """
        Websocket transport cannot use ML export operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/ml/export

        :raise CompatibilityError: on any use
        """
        message = "ML export is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/ml/export"
        raise CompatibilityError(full_message)

    def import_data(self, _path: Union[str, Path]):
        """
        Websocket transport cannot use import operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/import

        :raise CompatibilityError: on any use
        """
        message = "Import is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/import"
        raise CompatibilityError(full_message)

    def ml_import(self, _path: Union[str, Path]):
        """
        Websocket transport cannot use ML import operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/ml/import

        :raise CompatibilityError: on any use
        """
        message = "ML import is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/ml/import"
        raise CompatibilityError(full_message)

    def close(self):
        super().close()
        if self.is_connected():
            self._client.close()

    def is_connected(self) -> bool:
        return self._client.is_connected()

    def _run(self, data, callback: Callable = None) -> SurrealResult:
        result = self._client.send(data, callback)
        logger.info("Got result: %s", result)
        return result



from typing import Optional, Union

from surrealist.ql.statements.simple_statements import Where
from surrealist.ql.statements.common_statements import CanUseComment
from surrealist.ql.statements.statement import FinishedStatement, Statement


class PermissionsNone(FinishedStatement, CanUseComment):
    """
    Represents a `PERMISSIONS NONE` statement
    """
    def _clean_str(self):
        return f"{self._statement._clean_str()} PERMISSIONS NONE"


class PermissionsFull(FinishedStatement, CanUseComment):
    """
    Represents a `PERMISSIONS FULL` statement
    """
    def _clean_str(self):
        return f"{self._statement._clean_str()} PERMISSIONS FULL"


class PermissionsFor(FinishedStatement, CanUseComment):
    """
    Represents a `PERMISSIONS FOR` statement
    """
    def __init__(self, statement: Statement, select: Optional[Union[str, Where]] = None,
                 create: Optional[Union[str, Where]] = None,
                 update: Optional[Union[str, Where]] = None,
                 delete: Optional[Union[str, Where]] = None):
        super().__init__(statement)
        a_dict = {"select": select, "create": create, "update": update, "delete": delete}
        self._kw = {}
        for key, value in a_dict.items():
            if not value:
                continue
            if value not in self._kw:
                self._kw[value] = key
            else:
                self._kw[value] = f"{self._kw[value]}, {key}"
        self._kw = {v: k for k, v in self._kw.items()}

    def _clean_str(self):
        result = "\n FOR ".join(f"{k} {v}" for k, v in self._kw.items())
        return f"{self._statement._clean_str()} \nPERMISSIONS \n FOR {result}"


class CanUsePermissions:
    """
    Represents interface for using permissions
    """
    def permissions_none(self) -> PermissionsNone:
        """
        Represents PERMISSIONS NONE statement
        """
        return PermissionsNone(self)

    def permissions_full(self) -> PermissionsFull:
        """
        Represents PERMISSIONS FULL statement
        """
        return PermissionsFull(self)

    def permissions_for(self, **kwargs) -> PermissionsFor:
        """
        Represents PERMISSIONS FOR statement
        """
        return PermissionsFor(self, **kwargs)



from typing import Tuple

from surrealist.ql.statements.statement import Statement, FinishedStatement, IterableStatement
from surrealist.utils import OK


class Explain(FinishedStatement):
    """ Explain is not a child of IterableStatement because it has always only one result"""

    def __init__(self, statement: Statement, full: bool = False):
        super().__init__(statement)
        self._full = full

    def _clean_str(self):
        final = " FULL" if self._full else ""
        return f"{self._statement._clean_str()} EXPLAIN{final}"


class SelectUseExplain:

    def explain(self) -> Explain:
        return Explain(self, full=False)

    def explain_full(self) -> Explain:
        return Explain(self, full=True)


class Tempfiles(IterableStatement, SelectUseExplain):
    """
    Represent a TEMPFILES clause in CREATE statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} TEMPFILES"


class SelectUseTempfiles(SelectUseExplain):

    def tempfiles(self) -> Tempfiles:
        """
        Include TEMPFILES statement for the query
        When processing a large result set with many records, it is possible to use the TEMPFILES clause to specify that
        the statement should be processed in temporary files rather than memory. This significantly reduces memory
        usage, though it will also result in slower performance

        Refer to: https://surrealdb.com/docs/surrealql/statements/select#the-tempfiles-clause
        """
        return Tempfiles(self)


class Parallel(IterableStatement):
    """
    Represent a PARALLEL clause in CREATE statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} PARALLEL"


class SelectUseParallel(SelectUseTempfiles):
    def parallel(self) -> Parallel:
        return Parallel(self)


class Timeout(IterableStatement, SelectUseParallel):
    """
    Represent a TIMEOUT clause in CREATE statement.
    """

    def __init__(self, statement: Statement, duration: str):
        super().__init__(statement)
        self._duration = duration

    def _validate(self):
        if not any(self._duration.endswith(letter) for letter in ('ms', 's', 'm')):
            return f"Wrong duration {self._duration}, allowed postfix are (ms, s, m)"
        if ' ' in self._duration:
            return "Wrong duration format, should be like 5s"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} TIMEOUT {self._duration}"


class SelectUseTimeout(SelectUseParallel):
    def timeout(self, duration: str) -> Timeout:
        return Timeout(self, duration)


class Fetch(IterableStatement, SelectUseTimeout):
    """
    Represent a FETCH clause in CREATE statement.
    """

    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._args = args

    def _clean_str(self):
        add = "" if not self._args else ", ".join(self._args)
        if not add:
            return self._statement
        return f"{self._statement._clean_str()} FETCH {add}"


class SelectUseFetch(SelectUseTimeout):
    def fetch(self, *args) -> Fetch:
        return Fetch(self, *args)


class Start(IterableStatement, SelectUseFetch):
    """
    Represent a START clause in CREATE statement.
    """

    def __init__(self, statement: Statement, offset: int):
        super().__init__(statement)
        self._offset = offset

    def _validate(self):
        if self._offset < 0:
            return "Offset should not be negative number"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} START {self._offset}"


class SelectUseStart(SelectUseFetch):
    def start_at(self, offset: int) -> Start:
        return Start(self, offset)


class Limit(IterableStatement, SelectUseStart):
    """
    Represent a LIMIT clause in CREATE statement.
    """

    def __init__(self, statement: Statement, limit: int):
        super().__init__(statement)
        self._limit = limit

    def _validate(self):
        if self._limit < 1:
            return "Limit should not be less than 1"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} LIMIT {self._limit}"


class SelectUseLimit(SelectUseStart):
    def limit(self, limit: int) -> Limit:
        return Limit(self, limit)


class OrderByRand(IterableStatement, SelectUseLimit):
    """
    Represent ORDER BY RAND() clause in CREATE statement.
    """

    def __init__(self, statement: Statement):
        super().__init__(statement)
        self._statement = statement

    def _clean_str(self):
        return f"{self._statement._clean_str()} ORDER BY RAND()"


class OrderBy(IterableStatement, SelectUseLimit):
    """
    Represent a ORDER BY clause in CREATE statement.
    """

    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._statement = statement
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement
        what = ", ".join(self._args)
        return f"{self._statement._clean_str()} ORDER BY {what}"


class SelectUseOrder(SelectUseLimit):
    def order_by(self, *args: str) -> OrderBy:
        return OrderBy(self, *args)

    def order_by_rand(self) -> OrderByRand:
        return OrderByRand(self)


class GroupBy(IterableStatement, SelectUseOrder):
    """
    Represent a GROUP BY clause in CREATE statement.
    """

    def __init__(self, statement: Statement, *args):
        super().__init__(statement)
        self._statement = statement
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement
        what = ", ".join(self._args)
        return f"{self._statement._clean_str()} GROUP BY {what}"


class GroupAll(IterableStatement, SelectUseOrder):
    """
    Represent a GROUP ALL clause in CREATE statement.
    """

    def __init__(self, statement: Statement):
        super().__init__(statement)
        self._statement = statement

    def _clean_str(self):
        return f"{self._statement._clean_str()} GROUP ALL"


class SelectUseGroup(SelectUseOrder):
    def group_by(self, *args: str) -> GroupBy:
        return GroupBy(self, *args)

    def group_all(self) -> GroupAll:
        return GroupAll(self)


class Split(IterableStatement, SelectUseGroup):
    """
    Represent a SPLIT clause in CREATE statement.
    """

    def __init__(self, statement: Statement, field: str):
        super().__init__(statement)
        self._statement = statement
        self._field = field

    def _clean_str(self):
        return f"{self._statement._clean_str()} SPLIT {self._field}"


class SelectUseSplit(SelectUseGroup):
    def split(self, field: str) -> Split:
        return Split(self, field)


class Or(IterableStatement, SelectUseSplit):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._statement = statement
        self._predicate = predicate

    def AND(self, predicate):
        return And(self, predicate)

    def OR(self, predicate):
        return Or(self, predicate)

    def _clean_str(self):
        return f"{self._statement._clean_str()} OR {self._predicate}"


class And(IterableStatement, SelectUseSplit):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._statement = statement
        self._predicate = predicate

    def AND(self, predicate):
        return And(self, predicate)

    def OR(self, predicate):
        return Or(self, predicate)

    def _clean_str(self):
        return f"{self._statement._clean_str()} AND {self._predicate}"


class Where(IterableStatement, SelectUseSplit):
    """
    Represent a WHERE clause in CREATE statement.
    """

    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._statement = statement
        self._predicate = predicate

    def AND(self, predicate):
        return And(self, predicate)

    def OR(self, predicate):
        return Or(self, predicate)

    def _clean_str(self):
        return f"{self._statement._clean_str()} WHERE {self._predicate}"


class SelectUseWhere(SelectUseSplit):
    def where(self, predicate: str) -> Where:
        return Where(self, predicate)


class WithIndex(IterableStatement, SelectUseWhere):
    """
    Represent WITH INDEX clause in CREATE statement.
    """

    def __init__(self, statement: Statement, *index_names: str):
        super().__init__(statement)
        self._statement = statement
        self._index = index_names

    def _clean_str(self):
        if not self._index:
            return self._statement
        indexes = ", ".join(self._index)
        return f"{self._statement._clean_str()} WITH INDEX {indexes}"


class WithNoIndex(IterableStatement, SelectUseWhere):
    """
    Represent WITH NO INDEX clause in CREATE statement.
    """

    def __init__(self, statement: Statement):
        super().__init__(statement)
        self._statement = statement

    def _clean_str(self):
        return f"{self._statement._clean_str()} WITH NO INDEX"


class SelectUseIndex(SelectUseWhere):
    def with_index(self, *args: Tuple[str]) -> WithIndex:
        return WithIndex(self, *args)

    def with_no_index(self) -> WithNoIndex:
        return WithNoIndex(self)



from typing import List, Dict

from surrealist.connections import Connection
from surrealist.ql.statements.insert_statements import InsertUseDuplicate
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK, safe_dumps


class Insert(Statement, InsertUseDuplicate):
    """
    Represents INSERT INTO statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/insert

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_insert_examples.py

    INSERT [ IGNORE ] INTO @what
    [ @value
      | (@fields) VALUES (@values)
        [ ON DUPLICATE KEY UPDATE @field = @value ... ]
    ];
    """

    def __init__(self, connection: Connection, table_name: str, *args):
        """
        Init for Insert

        Notice: args can contain only one element if it is one of (List, Dict, Select) OR multiple tuples of the same
        length, representing names and values for insert

        Refer to: https://github.com/kotolex/surrealist/blob/master/examples/ql_insert_examples.py

        :param connection: underlying connection
        :param table_name: name of the table to operate on
        :param args: arguments
        """
        if not args:
            raise ValueError("Insert expects one or more arguments to insert")
        super().__init__(connection)
        self._table_name = table_name
        self._args = args

    def validate(self) -> List[str]:
        """
        Returns error if arguments are not appropriate for INSERT
        :return: list with OK or errors
        """
        args = self._args
        if len(args) > 1:
            key_len = len(args[0])
            for other in args[1:]:
                if len(other) != key_len:
                    return ["Length of names is not equal to one of the values tuple"]
        if len(args) == 1:
            if not isinstance(args[0], (List, Dict, Statement)):
                return ["If only one argument, it can be list, dict or Statement"]
        return [OK]

    def _clean_str(self):
        args = self._args
        if len(args) == 1:
            what = f"({args[0]._clean_str()})" if isinstance(args[0], Statement) else safe_dumps(args[0])
        else:
            names = f"({', '.join(args[0])})"
            data = ", ".join(safe_dumps(e) for e in args[1:])
            what = f"{names} VALUES {data}"
        return f"INSERT INTO {self._table_name} {what}"



from typing import Dict, List, Optional

from surrealist.ql.statements.common_statements import CanUseWhere
from surrealist.ql.statements.statement import FinishedStatement, Statement
from surrealist.ql.statements.utils import combine
from surrealist.utils import safe_dumps


class Set(FinishedStatement, CanUseWhere):
    """
    Represents SET part of the UPDATE statement
    SET @field = @value ...

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/update
    """
    def __init__(self, statement: Statement, result: Optional[str] = None, **kwargs):
        super().__init__(statement)
        self._kw = kwargs
        self._result = result

    def _clean_str(self):
        if not self._kw and not self._result:
            return self._statement._clean_str()
        args = combine(self._result, self._kw)
        return f"{self._statement._clean_str()} SET {args}"


class Patch(FinishedStatement, CanUseWhere):
    """
    Represents PATCH part of the UPDATE statement
    PATCH @value

    About patch: https://jsonpatch.com

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/update
    """
    def __init__(self, statement: Statement, operations: List[Dict]):
        super().__init__(statement)
        self._value = operations

    def _clean_str(self):
        return f"{self._statement._clean_str()} PATCH {safe_dumps(self._value)}"


class Merge(FinishedStatement, CanUseWhere):
    """
    Represents MERGE part of the UPDATE statement
    MERGE @value

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/update
    """
    def __init__(self, statement: Statement, value: Dict):
        super().__init__(statement)
        self._value = value

    def _clean_str(self):
        return f"{self._statement._clean_str()} MERGE {safe_dumps(self._value)}"


class Content(FinishedStatement, CanUseWhere):
    """
    Represents CONTENT part of the UPDATE statement
    CONTENT @value

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/update
    """
    def __init__(self, statement: Statement, value: Dict):
        super().__init__(statement)
        self._value = value

    def _clean_str(self):
        return f"{self._statement._clean_str()} CONTENT {safe_dumps(self._value)}"


class UpdateUseMethods(CanUseWhere):
    def set(self, result: Optional[str] = None, **kwargs) -> Set:
        """
        Sets the values for UPDATE statement.
        If both result and kwargs are used - will combine it, for example
        set("title ='Title'", text = "Text") will be 'SET title="Title", text="Text"'

        :param result: raw string to use for a complicated statement
        :param kwargs: simple key-values to use in statement
        :return: Set object
        """
        return Set(self, result, **kwargs)

    def content(self, value: Dict) -> Content:
        """
        Updates the record with given dictionary
        :param value: dict of data to update
        :return: Content oobject
        """
        return Content(self, value)

    def merge(self, value: Dict) -> Merge:
        """
        Merge (add) data to record
        :param value: dict with data to merge
        :return: Merge object
        """
        return Merge(self, value)

    def patch(self, operations: List[Dict]) -> Patch:
        """
        Patch (replace) data in record

        About patch format : https://jsonpatch.com

        :param operations: json-patch operations to update
        :return: Patch
        """
        return Patch(self, operations)



from typing import Optional
from surrealist.ql.statements.utils import combine


class Where:
    """
    Represents simple WHERE statement for DEFINE FIELD and DEFINE TABLE
    """
    def __init__(self, raw_string: Optional[str] = None, **kwargs):
        self._raw = raw_string
        self._kw = kwargs
        self._body = combine(self._raw, self._kw)
        self._body = f"WHERE {self._body}"

    def __str__(self):
        return self._body

    def __repr__(self):
        return f"Where({self._raw=}, {self._kw=})"

    def OR(self, raw_string: Optional[str] = None, **kwargs) -> "OrAnd":
        """
        Represents OR statement in WHERE
        """
        return OrAnd(self, True, raw_string, **kwargs)

    def AND(self, raw_string: Optional[str] = None, **kwargs) -> "OrAnd":
        """
        Represents AND statement in WHERE
        """
        return OrAnd(self, False, raw_string, **kwargs)


class OrAnd(Where):
    def __init__(self, parent: Where, use_or: bool = True, raw_string: Optional[str] = None, **kwargs):
        super().__init__(raw_string, **kwargs)
        self._or = use_or
        self._parent = parent
        self._cur_body = combine(raw_string, kwargs)
        self._type = "OR" if self._or else "AND"
        self._body = f"{parent} {self._type} {self._cur_body}"

    def __repr__(self):
        return f"{self._type}({self._raw=}, {self._kw=}, {self._parent:!r})"



from .connections import WebSocketConnection, HttpConnection, Connection
from .enums import Algorithm, AutoOrNone
from .errors import *
from .ql import Database, DatabaseConnectionsPool, Table, Where
from .record_id import RecordId
from .result import SurrealResult
from .surreal import Surreal
from .utils import get_uuid, to_surreal_datetime_str, to_datetime, LOG_FORMAT

__all__ = ("Surreal", "SurrealResult", "WebSocketConnection", "HttpConnection", "PySurrealError", "HttpConnectionError",
           "HttpClientError", "SurrealConnectionError", "WebSocketConnectionError", "WebSocketConnectionClosedError",
           "ConnectionParametersError", "CompatibilityError", "OperationOnClosedConnectionError", "WrongCallError",
           "Connection", "get_uuid", "Database", "Table", "Where", "DatabaseConnectionsPool", "AutoOrNone",
           "to_surreal_datetime_str", "to_datetime", "LOG_FORMAT", "Algorithm", "RecordId", "SurrealRecordIdError")



from .database import Database
from .pool_database import DatabaseConnectionsPool
from .statements.simple_statements import Where
from .table import Table

__all__ = ("Database", "DatabaseConnectionsPool", "Table", "Where",)



from typing import List

from .define import Define, Connection
from ...utils import OK

ALLOWED_LANGUAGES = ('Arabic', 'Danish', 'Dutch', 'English', 'French', 'German', 'Greek', 'Hungarian',
                     'Italian', 'Norwegian', 'Portuguese', 'Romanian', 'Russian', 'Spanish', 'Swedish',
                     'Tamil', 'Turkish')


class DefineAnalyzer(Define):
    """
    Represents DEFINE ANALYZER statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/analyzer

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE ANALYZER [ OVERWRITE | IF NOT EXISTS ] @name [ TOKENIZERS @tokenizers ] [ FILTERS @filters ]
    [ COMMENT @string ]

    """

    def __init__(self, connection: Connection, name: str):
        super().__init__(connection)
        self._name = name
        self._tokenizers = set()
        self._filters = set()
        self._error = None

    def if_not_exists(self) -> "DefineAnalyzer":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineAnalyzer":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def tokenizer_blank(self) -> "DefineAnalyzer":
        """
        Represents TOKENIZERS blank statement
        The blank tokenizer breaks down a text into tokens by creating a new token each time it encounters a space,
        tab, or newline character. It's a straightforward way to split text into words or chunks based on whitespace.
        For example, if you had the text "hello world", the blank tokenizer would create two tokens,
        ["hello" and "world"].

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#blank
        """
        self._tokenizers.add("blank")
        return self

    def tokenizer_camel(self) -> "DefineAnalyzer":
        """
        Represents TOKENIZERS camel statement
        The camel tokenizer is used for identifying and creating tokens when the next character in the text is
        uppercase. This is particularly useful for processing camelCase or PascalCase text, common in programming,
        to split them into meaningful words.
        For example, if you had the text "helloWorld", the camel tokenizer would create two tokens, ["hello", "World"].

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#camel
        """
        self._tokenizers.add("camel")
        return self

    def tokenizer_class(self) -> "DefineAnalyzer":
        """
        Represents TOKENIZERS class statement
        The class tokenizer segments text into tokens by detecting changes (digit, letter, punctuation, blank) in the
        Unicode class of characters. It creates a new token when the character class changes, distinguishing between
        digits, letters, punctuation, and blanks. This allows for flexible tokenization based on character types.
        For example, if you had the text "123abc!XYZ", the class tokenizer would create four tokens,
        ["123", "abc", "!", "XYZ"].

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#class
        """
        self._tokenizers.add("class")
        return self

    def tokenizer_punct(self) -> "DefineAnalyzer":
        """
        Represents TOKENIZERS punct statement
        The punct tokenizer generates tokens by breaking the text whenever a punctuation character is encountered.
        It's suitable for tokenizing sentences or breaking text into smaller units based on punctuation marks.
        For example, if you had the text "Hello, World!", the punct tokenizer would create four tokens,
        ["Hello", ",", "World", "!"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#punct
        """
        self._tokenizers.add("punct")
        return self

    def filter_ascii(self) -> "DefineAnalyzer":
        """
        Represents FILTERS ascii statement
        The ascii filter is responsible for processing tokens by replacing or removing diacritical marks (accents and
        special characters) from the text. It helps standardize text by converting accented characters to their basic
        ASCII equivalents, making it more suitable for various text analysis tasks.
        For example, if you had the text "résumé café", the ascii filter would create two tokens, ["resume", "cafe"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#ascii
        """
        self._filters.add("ascii")
        return self

    def filter_lowercase(self) -> "DefineAnalyzer":
        """
        Represents FILTERS lowercase statement
        The lowercase filter converts tokens to lowercase, ensuring that text is consistently in lowercase format.
        This is often used to make text case-insensitive for search and analysis purposes.
        For example, if you had the text "Hello World", the lowercase filter would create two tokens, ["hello", "world"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#lowercase
        """
        self._filters.add("lowercase")
        return self

    def filter_uppercase(self) -> "DefineAnalyzer":
        """
        Represents FILTERS uppercase statement
        The uppercase filter converts tokens to uppercase, ensuring text consistency in uppercase format. It can be
        useful when case-insensitivity is required for specific analysis or search operations.
        For example, if you had the text "Hello World", the uppercase filter would create two tokens, ["HELLO", "WORLD"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#uppercase
        """
        self._filters.add("uppercase")
        return self

    def filter_ngram(self, minimum: int, maximum: int) -> "DefineAnalyzer":
        """
        Represents FILTERS ngram(min,max) statement
        The ngram filter is used to create a sequence of 'n' tokens from a given sample of text or speech. These items
        can be syllables, letters, words or base pairs according to the application. It accepts two parameters min and
        max which indicates that you want to create n-grams starting from min to size of max.
        For example, if you had the text "apple banana", the ngram filter would create these tokens:
        ["a", "p", "p", "l", "e", "ap", "pp", "pl", "le", "app", "ppl", "ple", "b", "a", "n", "a", "n", "a", "ba", "an",
         "na", "an", "na", "ban", "ana", "nan", "ana"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#ngramminmax
        """
        self._filters.add(f"ngram({minimum},{maximum})")
        return self

    def filter_edgengram(self, minimum: int, maximum: int) -> "DefineAnalyzer":
        """
        Represents FILTERS edgengram(min,max) statement
        The edgengram filter is used to create tokens that represent prefixes of terms. It generates a sequence of
        tokens that gradually build up a term, which can be useful for autocomplete or searching based on partial words.
        It accepts two parameters min and max which define the minimum and maximum amount of characters in the prefix.
        For example, if you had the text "apple banana", the edgengram filter would create six tokens,
        ["a", "ap", "app", "b", "ba", "ban"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#edgengramminmax
        """
        self._filters.add(f"edgengram({minimum},{maximum})")
        return self

    def filter_snowball(self, language: str) -> "DefineAnalyzer":
        """
        Represents FILTERS snowball(language) statement
        The snowball filter applies Snowball stemming to tokens, reducing them to their root form and converts the case
        to lowercase. The following supported languages can be passed as a parameter in snowball: Arabic, Danish, Dutch,
        English, French, German, Greek, Hungarian, Italian, Norwegian, Portuguese, Romanian, Russian, Spanish, Swedish,
        Tamil, Turkish.
        For example, if you had the text "running cats", the snowball filter would create two tokens, ["run", "cat"]

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/analyzer#snowballlanguage
        """
        if not any(language.lower() == lang.lower() for lang in ALLOWED_LANGUAGES):
            self._error = language
        self._filters.add(f"snowball({language})")
        return self

    def validate(self) -> List[str]:
        if self._error:
            return [f"{self._error} is not in {ALLOWED_LANGUAGES}"]
        return [OK]

    def _clean_str(self):
        tok = "" if not self._tokenizers else f" TOKENIZERS {', '.join(sorted(self._tokenizers))}"
        filters = "" if not self._filters else f" FILTERS {', '.join(sorted(self._filters))}"
        return f"DEFINE ANALYZER{self._exists()} {self._name}{tok}{filters}{self._comment()}"



from typing import List, Optional

from surrealist.connections import Connection
from surrealist.utils import OK, StrOrInt
from .statement import Statement


class Show(Statement):
    """
    Represents SHOW CHANGES statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/show

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_show_examples.py

    SHOW CHANGES FOR TABLE @tablename SINCE @timestamp | @versionstamp [ LIMIT @number ]

    """

    def __init__(self, connection: Connection, table_name: str, since: Optional[StrOrInt] = None):
        super().__init__(connection)
        self._table_name = table_name
        self._since = since
        self._limit = None

    def validate(self) -> List[str]:
        result = []
        if self._since and isinstance(self._since, str):
            if not self._since.startswith("d"):
                result.append("Timestamp in the wrong format, you need iso-date like d'2024-01-01T10:10:10.000001Z'")
        if self._limit and self._limit < 1:
            result.append("Limit should not be less than 1")
        return [OK] if not result else result

    def limit(self, limit: int) -> "Show":
        """
        Limits number of results to show (LIMIT statement)
        :param limit: number of results
        :return: Show object
        """
        self._limit = limit
        return self

    def since(self, datetime_or_versionstamp: [StrOrInt]) -> "Show":
        """
        Init timestamp or versionstamp since is to show updates,
        it should be a surreal timestamp like d'2024-01-01T10:10:10.000001Z' or an integer versionstamp

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/show#basic-usage

        :param datetime_or_versionstamp: surreal timestamp or versionstamp
        :return: Show object
        """
        self._since = datetime_or_versionstamp
        return self

    def _clean_str(self):
        if not self._since:
            self._since = 1 # default value, shows all changes
        limit = f" LIMIT {self._limit}" if self._limit else ""
        return f'SHOW CHANGES FOR TABLE {self._table_name} SINCE {self._since}{limit}'



import logging
from os import cpu_count
from typing import Optional, Tuple

from surrealist.connections.pool import Pool
from surrealist.ql.database import Database
from surrealist.utils import DEFAULT_TIMEOUT

CORES_COUNT = cpu_count()
logger = logging.getLogger("surrealist.databaseQL.pool")


class DatabaseConnectionsPool(Database):
    """
    Represents a pool of connections to a database, for using in high-load multithreading environments. As a child of
    Database object it can be used everywhere where Database can.
    The main difference is a number of created connections. By default, the minimum number is equal to CPU cores count
    in the system and cannot be less than 2.

    During work, if no more non-busy connections to use, then new connection will be created until its number not
    reaches maximum for the pool

    Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#connections-pool

    """

    def __init__(self, url: str, namespace: str, database: str, access: Optional[str] = None,
                 credentials: Optional[Tuple[str, str]] = None,
                 use_http: bool = False, timeout: int = DEFAULT_TIMEOUT,
                 min_connections: int = CORES_COUNT, max_connections: int = 50):
        """
        All parameters are the same as for Surreal or Database object

        :param min_connections: minimum number of connections, it cannot be less than 2
        :param max_connections: maximum number of connections, it cannot be more than 50
        """
        self._options = {
            "url": url, "namespace": namespace, "database": database, "access": access, "credentials": credentials,
            "use_http": use_http, "timeout": timeout, "min_connections": min_connections,
            "max_connections": max_connections
        }
        super().__init__(url, namespace, database, access, credentials, use_http, timeout)
        self._connection = Pool(self._connection, **self._options)
        self._connected = True
        self._min = min_connections
        self._max = max_connections
        logger.info("Pool DatabaseQL is up")

    @property
    def connections_count(self) -> int:
        """
        Get the current number of connections, all of them can be busy now

        :return: number of connections in the pool
        """
        return self._connection.connections_count

    @property
    def min_connections(self) -> int:
        """
        Returns minimum number of connections for current pool

        :return: minimum number of connections in the pool
        """
        return self._min

    @property
    def max_connections(self) -> int:
        """
        Returns maximum number of connections for current pool

        :return: maximum number of connections in the pool
        """
        return self._max

    def __repr__(self):
        return f"DatabasePool(namespace={self._namespace}, name={self._database}, connected={self.is_connected()}," \
               f"connections_count={self.connections_count}, min_connections={self._min}, max_connections={self._max})"



import json
from abc import ABC
from typing import List, Union, Any, Optional, Tuple

from surrealist.connections import Connection
from surrealist.enums import Algorithm
from surrealist.ql.statements.define_index_statements import CanUseIndexTypes, CanUseConcurrently
from surrealist.ql.statements.permissions import CanUsePermissions
from surrealist.ql.statements.statement import Statement, FinishedStatement
from surrealist.utils import OK


class Define(Statement, ABC):
    """
    Parent for all DEFINE statements, can use IF NOT EXISTS statement
    """

    def __init__(self, connection: Connection):
        super().__init__(connection)
        self._if_not_exists = None
        self._text = None

    def _exists(self) -> str:
        if self._if_not_exists is None:
            return ""
        return " IF NOT EXISTS" if self._if_not_exists else " OVERWRITE"

    def _comment(self) -> str:
        return f" COMMENT {json.dumps(self._text)}" if self._text else ""

    def if_not_exists(self) -> "Define":
        """
        Adds IF NOT EXISTS statement to the query
        :return: self
        """
        self._if_not_exists = True
        return self

    def overwrite(self) -> "Define":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def comment(self, comment: str) -> FinishedStatement:
        """
        Adds COMMENT statement to the query
        :param comment: comment text
        :return: self
        """
        self._text = comment
        return FinishedStatement(self)


class DefineEvent(Define):
    """
    Represents DEFINE EVENT statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/event

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE EVENT [ OVERWRITE | IF NOT EXISTS ] @name ON [ TABLE ] @table [ WHEN @expression ] THEN @expression
    [ COMMENT @string ]

    """

    def __init__(self, connection: Connection, name: str, table_name: str, then: Union[str, Statement]):
        super().__init__(connection)
        self._name = name
        self._table_name = table_name
        self._when = None
        self._then = then if not isinstance(then, Statement) else f"({then._clean_str()})"

    def if_not_exists(self) -> "DefineEvent":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineEvent":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def when(self, predicate: str) -> "DefineEvent":
        """
        Represents condition for event, WHEN statement

        :param predicate: condition for event
        """
        self._when = predicate
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        when = f"WHEN {self._when} " if self._when else ""
        return f"DEFINE EVENT{self._exists()} {self._name} ON TABLE {self._table_name} {when}" \
               f"THEN {self._then}{self._comment()}"


class DefineParam(Define):
    """
    Represents DEFINE PARAM statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/param

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE PARAM [ OVERWRITE | IF NOT EXISTS ] $@name VALUE @value [ COMMENT @string ]
    """

    def __init__(self, connection: Connection, param_name: str, value: Any):
        super().__init__(connection)
        self._name = param_name
        self._value = value

    def if_not_exists(self) -> "DefineParam":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineParam":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        return f"DEFINE PARAM{self._exists()} ${self._name} VALUE {self._value}{self._comment()}"


class DefineScope(Define):
    """
    Deprecated since SurrealDB 2.x, use define_access_record instead!

    Represents DEFINE SCOPE statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE SCOPE [ OVERWRITE | IF NOT EXISTS ] @name SESSION @duration SIGNUP @expression SIGNIN @expression
    [ COMMENT @string ]
    """

    def __init__(self, connection: Connection, name: str, duration: str, signup: Union[str, Statement],
                 signin: Union[str, Statement]):
        super().__init__(connection)
        self._name = name
        self._duration = duration
        self._signup = signup if not isinstance(signup, Statement) else f"({signup._clean_str()})"
        self._signin = signin if not isinstance(signin, Statement) else f"({signin._clean_str()})"

    def if_not_exists(self) -> "DefineScope":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineScope":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def validate(self) -> List[str]:
        if ' ' in self._duration:
            return ["Wrong duration format, should be like 24h"]
        return [OK]

    def _clean_str(self):
        return f"DEFINE SCOPE{self._exists()} {self._name} SESSION {self._duration} \nSIGNUP {self._signup} " \
               f"\nSIGNIN {self._signin}{self._comment()}"


class DefineIndex(Define, CanUseIndexTypes, CanUseConcurrently):
    """
    Represents DEFINE INDEX statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/indexes

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE INDEX [ OVERWRITE | IF NOT EXISTS ] @name ON [ TABLE ] @table [ FIELDS | COLUMNS ] @fields
    [ UNIQUE
        | SEARCH ANALYZER @analyzer [ BM25 [(@k1, @b)] ] [ HIGHLIGHTS ]
        | MTREE DIMENSION @dimension [ TYPE @type ] [ DIST @distance ] [ CAPACITY @capacity]
        | HNSW DIMENSION @dimension [ TYPE @type ] [DIST @distance] [ EFC @efc ] [ M @m ]
    ]
    [ COMMENT @string ]
    [ CONCURRENTLY ]
    """

    def __init__(self, connection: Connection, name: str, table_name: str):
        super().__init__(connection)
        self._name = name
        self._table_name = table_name
        self._fields = None
        self._uni = False
        self._analyzer = None

    def if_not_exists(self) -> "DefineIndex":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineIndex":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def fields(self, fields: str) -> "DefineIndex":
        """
        Represents fields for index, FIELDS statement

        :param fields: fields to index
        """
        self._fields = f"FIELDS {fields}"
        return self

    def columns(self, columns: str) -> "DefineIndex":
        """
        Represents fields for index, COLUMNS statement

        :param columns: fields to index
        """
        self._fields = f"COLUMNS {columns}"
        return self

    def validate(self) -> List[str]:
        if not self._fields:
            return ["You need to specify FIELDS or COLUMNS"]
        return [OK]

    def _clean_str(self):
        return f"DEFINE INDEX{self._exists()} {self._name} ON TABLE {self._table_name} {self._fields}{self._comment()}"


class DefineToken(Define):
    """
    Deprecated since SurrealDB 2.x, use define_access_jwt or define_access_record instead!

    Represents DEFINE TOKEN statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/token

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE TOKEN [ OVERWRITE | IF NOT EXISTS ] @name ON [ NAMESPACE | DATABASE | SCOPE @scope ] TYPE @type VALUE @value
    [ COMMENT @string ]
    """

    def __init__(self, connection: Connection, name: str, token_type: Algorithm, value: str):
        super().__init__(connection)
        self._name = name
        self._type = token_type
        self._value = value

    def if_not_exists(self) -> "DefineToken":
        self._if_not_exists = True
        return self

    def validate(self) -> List[str]:
        if not isinstance(self._type, Algorithm):
            return ["Invalid token type, you should use one of Algorithm enumerations"]
        return [OK]

    def _clean_str(self):
        return f'DEFINE TOKEN{self._exists()} {self._name} ON DATABASE \nTYPE {self._type.name} \n' \
               f'VALUE "{self._value}"{self._comment()}'


class DefineTable(Define, CanUsePermissions):
    """
    Represents DEFINE TABLE statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/table

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    
    DEFINE TABLE [ OVERWRITE | IF NOT EXISTS ] @name
    [ DROP ]
    [ SCHEMAFULL | SCHEMALESS ]
    [ TYPE [ ANY | NORMAL | RELATION [ IN | FROM ] @table [ OUT | TO ] @table ] [ ENFORCED ]]
    [ AS SELECT @projections
        FROM @tables
        [ WHERE @condition ]
        [ GROUP [ BY ] @groups ]
    ]
    [CHANGEFEED @duration [INCLUDE ORIGINAL] ]
    [ PERMISSIONS [ NONE | FULL
        | FOR select @expression
        | FOR create @expression
        | FOR update @expression
        | FOR delete @expression
    ] ]
    [ COMMENT @string ]
    """

    def __init__(self, connection: Connection, name: str):
        super().__init__(connection)
        self._name = name
        self._drop = False
        self._less = False
        self._full = False
        self._alias = None
        self._changefeed = None
        self._type = None

    def if_not_exists(self) -> "DefineTable":
        """
        Represents IF NOT EXISTS statement
        """
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineTable":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def drop(self) -> "DefineTable":
        """
        Represents DROP statement
        """
        self._drop = True
        return self

    def schemafull(self) -> "DefineTable":
        """
        Represents SCHEMAFULL statement
        """
        self._full = True
        self._less = False
        return self

    def schemaless(self) -> "DefineTable":
        """
        Represents SCHEMALESS statement
        """
        self._full = False
        self._less = True
        return self

    def changefeed(self, duration: str, include_original: bool = False) -> "DefineTable":
        """
        Represents CHANGEFEED statement

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/table#using-changefeed-clause

        :param duration: valid string representation for duration, for example, "1s"
        :param include_original: if True, then add INCLUDE ORIGINAL statement
        """
        self._changefeed = (duration, include_original)
        return self

    def alias(self, select: Union[str, Statement]) -> "DefineTable":
        """
        Represents AS statements, Select statement or raw string expected
        :param select: Select statement or string
        """
        if isinstance(select, Statement):
            self._alias = select._clean_str()
        else:
            self._alias = select
        return self

    def type_any(self) -> "DefineTable":
        """
        Represents TYPE ANY statement
        """
        self._type = "ANY"
        return self

    def type_normal(self) -> "DefineTable":
        """
        Represents TYPE NORMAL statement
        """
        self._type = "NORMAL"
        return self

    def type_relation(self, from_to: Optional[Tuple] = None, use_from_to: bool = True,
                      enforced: bool = False) -> "DefineTable":
        """
        Represents TYPE RELATE statement

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/table#table-with-specialized-type-clause-since-140

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param from_to: optional pair of source and target table, both must be provided (not None)
        :param use_from_to: if True, FROM TO syntax will be used, if False, IN OUT syntax will be used
        :param enforced: if True, adds ENFORCED statement to query
        """
        if from_to and from_to[0] and from_to[1]:
            if use_from_to:
                self._type = f"RELATION FROM {from_to[0]} TO {from_to[1]}"
            else:
                self._type = f"RELATION IN {from_to[0]} OUT {from_to[1]}"
        else:
            self._type = "RELATION"
        if enforced:
            self._type = f"{self._type} ENFORCED"
        return self

    def validate(self) -> List[str]:
        durations = ('w', 'y', 'd', 'h', 'ms', 's', 'm')
        if self._changefeed:
            actual_duration, _ = self._changefeed
            if not any(actual_duration.endswith(letter) for letter in durations):
                return [f"Wrong duration {actual_duration}, allowed postfix are {durations}"]
        return [OK]

    def _clean_str(self):
        drop = "" if not self._drop else " DROP"
        schema = ""
        if self._less:
            schema = " SCHEMALESS"
        elif self._full:
            schema = " SCHEMAFULL"
        alias = "" if not self._alias else f" AS\n {self._alias}\n"
        feed = ""
        if self._changefeed:
            feed = f" CHANGEFEED {self._changefeed[0]}"
            if self._changefeed[1]:
                feed = f"{feed} INCLUDE ORIGINAL"
        type_ = "" if not self._type else f" TYPE {self._type}"
        return f'DEFINE TABLE{self._exists()} {self._name}{drop}{schema}{type_}{alias}{feed}{self._comment()}'


class DefineField(Define, CanUsePermissions):
    """
    Represents DEFINE FIELD statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/field

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/define_field.py

    DEFINE FIELD [ OVERWRITE | IF NOT EXISTS ] @name ON [ TABLE ] @table
    [ [ FLEXIBLE ] TYPE @type ]
    [ DEFAULT @expression ]
  [ READONLY ]
    [ VALUE @expression ]
    [ ASSERT @expression ]
    [ PERMISSIONS [ NONE | FULL
        | FOR select @expression
        | FOR create @expression
        | FOR update @expression
        | FOR delete @expression
    ] ]
  [ COMMENT @string ]
    """

    def __init__(self, connection: Connection, field_name: str, table_name: str):
        super().__init__(connection)
        self._field_name = field_name
        self._table_name = table_name
        self._flex = False
        self._type = None
        self._default = None
        self._readonly = False
        self._value = None
        self._assert = None

    def if_not_exists(self) -> "DefineField":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineField":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def type(self, type_name: str, is_flexible: bool = False) -> "DefineField":
        """
        Represents TYPE statement and FLEXIBLE statement
        """
        self._flex = is_flexible
        self._type = type_name
        return self

    def default(self, value: str) -> "DefineField":
        """
        Represents DEFAULT statement
        """
        self._default = value
        return self

    def read_only(self) -> "DefineField":
        """
        Represents READONLY statement
        """
        self._readonly = True
        return self

    def value(self, value: str) -> "DefineField":
        """
        Represents VALUE statement
        """
        self._value = value
        return self

    def asserts(self, value: str) -> "DefineField":
        """
        Represents ASSERT statement
        """
        self._assert = value
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        type_ = "" if not self._type else f" TYPE {self._type}"
        if type_ and self._flex:
            type_ = f" FLEXIBLE{type_}"
        default = "" if not self._default else f" DEFAULT {self._default}"
        ro = "" if not self._readonly else " READONLY"
        value = "" if not self._value else f" VALUE {self._value}"
        asserts = "" if not self._assert else f" ASSERT {self._assert}"
        return f'DEFINE FIELD{self._exists()} {self._field_name} ' \
               f'ON TABLE {self._table_name}{type_}{default}{ro}{value}{asserts}{self._comment()}'



from typing import List, Optional

from surrealist.connections import Connection
from surrealist.ql.statements.statement import Statement
from surrealist.ql.statements.update_statements import UpdateUseMethods
from surrealist.utils import OK, get_table_or_record_id


class Update(Statement, UpdateUseMethods):
    """
    Represents UPDATE statement, it should be able to use any statements from documentation
    The UPDATE statement can be used to update existing records in the database.
    If the record does not exist, the statement will fail and no records will be updated.

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_update_examples.py

    UPDATE [ ONLY ] @targets
    [ CONTENT @value
      | MERGE @value
      | PATCH @value
      | SET @field = @value ...
    ]
    [ WHERE @condition ]
    [ RETURN NONE | RETURN BEFORE | RETURN AFTER | RETURN DIFF | RETURN @statement_param, ... ]
    [ TIMEOUT @duration ]
    [ PARALLEL ];
    """

    def __init__(self, connection: Connection, table_name: str, record_id: Optional[str] = None):
        super().__init__(connection)
        self._table_name = table_name
        self._only = False
        self._record_id = record_id
        self._name = get_table_or_record_id(table_name, record_id)

    def validate(self) -> List[str]:
        return [OK]

    def only(self) -> "Update":
        """
        Include ONLY statement for the query
        """
        self._only = True
        return self

    def _clean_str(self):
        what = "" if not self._only else " ONLY"
        return f"UPDATE{what} {self._name}"



from typing import List, Dict, Optional

from surrealist.connections.connection import Connection
from surrealist.utils import OK
from .define import Define


class DefineUser(Define):
    """
    Represents DEFINE USER ON DATABASE statement, a default role is VIEWER

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/user

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    DEFINE USER [ OVERWRITE | IF NOT EXISTS ] @name
    ON [ ROOT | NAMESPACE | DATABASE ]
    [ PASSWORD @pass | PASSHASH @hash ]
    [ ROLES @roles ]
    [ DURATION [ FOR TOKEN @duration [ , ] ] [ FOR SESSION @duration ] ]
    [ COMMENT @string ]
    """

    def __init__(self, connection: Connection, user_name: str):
        super().__init__(connection)
        self._name = user_name
        self._role = None
        self._pass = None
        self._hash = None
        self._durations: Dict[str, Optional[str]] = {"token": None, "session": None}

    def if_not_exists(self) -> "DefineUser":
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineUser":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def validate(self) -> List[str]:
        for duration in self._durations.values():
            if not duration:
                continue
            if not any(duration.endswith(letter) for letter in ('d', 'h', 's', 'm')):
                return [f"Wrong duration {duration}, allowed postfix are ('d', 'h', 's', 'm')"]
            if ' ' in duration:
                return ["Wrong duration format, should be like 5h"]
        return [OK]

    def password(self, value: str) -> "DefineUser":
        """
        Adds PASSWORD clause to a final statement
        """
        self._pass = value
        self._hash = None
        return self

    def passhash(self, value: str) -> "DefineUser":
        """
        Adds PASSHASH clause to a final statement
        """
        self._pass = None
        self._hash = value
        return self

    def role_owner(self) -> "DefineUser":
        """
        Represents role OWNER

        """
        self._role = "OWNER"
        return self

    def role_editor(self) -> "DefineUser":
        """
        Represents role EDITOR

        """
        self._role = "EDITOR"
        return self

    def role_viewer(self) -> "DefineUser":
        """
        Represents role VIEWER

        """
        self._role = "VIEWER"
        return self

    def duration_token(self, duration: str) -> "DefineUser":
        """
        Adds DURATION FOR TOKEN clause to a final statement
        """
        self._durations["token"] = duration
        return self

    def duration_session(self, duration: str) -> "DefineUser":
        """
        Adds DURATION FOR SESSION clause to a final statement
        """
        self._durations["session"] = duration
        return self

    def _clean_str(self):
        a_list = []
        if self._durations["token"]:
            a_list.append(f"DURATION FOR TOKEN {self._durations['token']}")
        if self._durations["session"]:
            a_list.append(f"DURATION FOR SESSION {self._durations['session']}")
        durations = ", ".join(a_list)
        if durations:
            durations = f" {durations}"
        pass_ = ""
        if self._pass:
            pass_ = f" PASSWORD '{self._pass}'"
        if self._hash:
            pass_ = f" PASSHASH '{self._hash}'"
        roles = "" if not self._role else f" ROLES {self._role}"
        return f"DEFINE USER{self._exists()} {self._name} ON DATABASE{pass_}{roles}{durations}{self._comment()}"



from typing import Optional, List

from surrealist.connections.connection import Connection
from surrealist.enums import AutoOrNone
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class DefineConfig(Statement):
    """
    The DEFINE CONFIG GRAPHQL statement allows you to configure how your database’s tables and functions are exposed
    via the GraphQL API. This configuration is essential for enabling GraphQL functionality in your database,
    specifying which tables and functions should be included or excluded from the GraphQL schema.

    Refer to: https://surrealdb.com/docs/surrealql/statements/define/config

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/define_config.py

    DEFINE CONFIG [OVERWRITE | IF NOT EXISTS] GRAPHQL [ AUTO | NONE]
    [ TABLES (AUTO | NONE | INCLUDE table1, table2, ...) ]
    [ FUNCTIONS (AUTO | NONE | INCLUDE [function1, function2, ...] | EXCLUDE [function1, function2, ...]) ]
    """

    def validate(self) -> List[str]:
        return [OK]

    def __init__(self, connection: Connection, kind: Optional[AutoOrNone] = None):
        super().__init__(connection)
        self._if_not_exists = None
        self._auto = kind
        self._tables = None
        self._functions = None

    def if_not_exists(self) -> "DefineConfig":
        """
        Represents the IF NOT EXISTS statement.
        :return: DefineConfig object
        """
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineConfig":
        """
        Adds OVERWRITE statement to the query
        :return: DefineConfig object
        """
        self._if_not_exists = False
        return self

    def tables_kind(self, kind: AutoOrNone) -> "DefineConfig":
        """
        Represents the TABLES AUTO | NONE statement.
        :return: DefineConfig object
        """
        self._tables = kind
        return self

    def tables_include(self, raw_tables: str) -> "DefineConfig":
        """
        Represents the TABLES INCLUDE table1, table2, ... statement.
        :param raw_tables: raw string representation of tables, will be used as is
        :return: DefineConfig object
        """
        self._tables = f"INCLUDE {raw_tables}"
        return self

    def functions_kind(self, kind: AutoOrNone) -> "DefineConfig":
        """
        Represents the FUNCTIONS AUTO | NONE statement.
        :return: DefineConfig object
        """
        self._functions = kind
        return self

    def functions_include(self, raw_functions: str) -> "DefineConfig":
        """
        Represents the FUNCTIONS INCLUDE [function1, function2, ...] statement.
        :param raw_functions: raw string representation of functions, will be used as is
        :return: DefineConfig object
        """
        self._functions = f"INCLUDE {raw_functions}"
        return self

    def functions_exclude(self, raw_functions: str) -> "DefineConfig":
        """
        Represents the FUNCTIONS EXCLUDE [function1, function2, ...] statement.
        :param raw_functions: raw string representation of functions, will be used as is
        :return: DefineConfig object
        """
        self._functions = f"EXCLUDE {raw_functions}"
        return self

    def _clean_str(self):
        exists = ""
        if self._if_not_exists:
            exists = " IF NOT EXISTS"
        elif self._if_not_exists is False:
            exists = " OVERWRITE"
        auto = ""
        if self._auto == AutoOrNone.AUTO:
            auto = " AUTO"
        elif self._auto == AutoOrNone.NONE:
            auto = " NONE"
        tables = ""
        if self._tables is not None:
            if self._tables == AutoOrNone.AUTO:
                tables = " TABLES AUTO"
            elif self._tables == AutoOrNone.NONE:
                tables = " TABLES NONE"
            else:
                tables = f" TABLES {self._tables}"
        functions = ""
        if self._functions is not None:
            if self._functions == AutoOrNone.AUTO:
                functions = " FUNCTIONS AUTO"
            elif self._functions == AutoOrNone.NONE:
                functions = " FUNCTIONS NONE"
            else:
                functions = f" FUNCTIONS {self._functions}"

        return f"DEFINE CONFIG{exists} GRAPHQL{auto}{tables}{functions}"



import urllib.parse
from logging import getLogger
from typing import Tuple, Optional

from surrealist.clients import HttpClient
from surrealist.connections.connection import Connection
from surrealist.connections.http_connection import HttpConnection
from surrealist.connections.ws_connection import WebSocketConnection
from surrealist.errors import HttpClientError, SurrealConnectionError, ConnectionParametersError
from surrealist.utils import DEFAULT_TIMEOUT, ENCODING, OK, HTTP_OK, NS, DB, AC

logger = getLogger("surrealist")


class Surreal:
    """
    Represents the SurrealDB client, all connections should be established via this class. By default, connection will
    use websocket transport (recommended). This class uses logging to monitor all actions, can check a version and
    state of the SurrealDb server.
    If you have only one SurrealDB server, you need exactly one object of this class!

    Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#connect-to-surrealdb
    """

    def __init__(self, url: str, namespace: Optional[str] = None, database: Optional[str] = None,
                 access: Optional[str] = None, credentials: Tuple[str, str] = None, use_http: bool = False,
                 timeout: int = DEFAULT_TIMEOUT):
        """
        Initiating all parameters for connection, this method does not check or validates anything by itself, just save
        data for future use. To make sure your url is valid and accessible - use **is_ready** method of Surreal object.

        About debug mode: https://github.com/kotolex/surrealist?tab=readme-ov-file#logging-and-debug-mode

        :param url: database url, for local database http://127.0.0.1:8000/
        :param namespace: namespace to use
        :param database: a database to use, make sure to use both namespace and database or none of them
        :param access: access method, this parameter is only required if you are signing in using an access method,
        otherwise you can omit it for database, namespace and root users.
        :param credentials: pair of user and password, for example ("root", "root")
        :param use_http: boolean flag of using http transport. Will use websocket-client if False.
        It is strongly recommended to use websocket transport as it is more powerful.
        :param timeout: connection timeout in seconds
        """
        self._client = HttpConnection if use_http else WebSocketConnection
        self.db_params = {}
        if namespace:
            self.db_params[NS] = namespace
        if database:
            if not namespace:
                msg = "Database can't be used without a namespace. Please specify namespace to use"
                logger.error(msg)
                raise ConnectionParametersError(msg)
            self.db_params[DB] = database
        if access:
            if not database:
                msg = "Access method can't be used without a namespace and database. Please specify it"
                logger.error(msg)
                raise ConnectionParametersError(msg)
            self.db_params[AC] = access
        if not self.db_params:
            self.db_params = None
        self._url, self._possible_url, self._is_http_url = url, url, True
        self.set_url(url)
        self.credentials = credentials
        self.timeout = timeout

    def set_url(self, url: str):
        """
        Setting base and predicted urls to work with. If url starts with http, then both url and predicted url will be
        the same. If url in ws(s) format, method will try to create a predicted url based on it.

        For example for wss://127.0.0.1:9000/some/rps predicted http url will be https://127.0.0.1:9000/

        :param url: url of the SurrealDb server
        """
        self._url = url
        self._is_http_url = url.startswith("http")
        # we should change only http endpoints
        if self._is_http_url and not url.endswith("/"):
            self._url = f"{url}/"
        self._possible_url = self._url
        # we can try to predict http url when wss url was specified
        if not self._is_http_url:
            _url = urllib.parse.urlparse(url.lower())
            self._possible_url = f"{_url.scheme.replace('ws', 'http')}://{_url.netloc}/"
            logger.info("Predicted url is: %s", self._possible_url)

    def connect(self) -> Connection:
        """
        Actually connects to a database via chosen transport (websocket or http), uses specified namespace, database and
        credentials parameters, so can raise exception if connect will fail. If this method succeeded, you are
        ready to go with SurrealDB

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/connect.py

        :return: connection object to work with SurrealDB
        :raise SurrealConnectionError: if cant connect with specified parameters
        """
        return self._client(self._url, db_params=self.db_params, credentials=self.credentials, timeout=self.timeout)

    def is_ready(self) -> bool:
        """
        Checks that SurrealDB server is up and running. Under the hood it calls **health** and **status** methods to
        make sure database web server is running, database server and storage engine are running.

        Notice: Method uses the specified url, but not namespace, database or credentials, so success of this method
        (return True) should not be interpreted as authorization success

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#status

        https://docs.surrealdb.com/docs/integration/http#health

        :return: True if you can use this server, False otherwise
        :raise SurrealConnectionError: if cant connect to http-endpoint
        """
        return self.status() == OK and self.health() == OK

    def status(self) -> str:
        """
        Checks that SurrealDB web server is running.

        Refer to: https://docs.surrealdb.com/docs/integration/http#status

        :return: "OK" if everything is ok, or error text
        :raise SurrealConnectionError: if cant connect to http-endpoint
        """
        code, text = self.__get("status")
        return OK if code == HTTP_OK else text

    def health(self) -> str:
        """
        Checks that SurrealDB server and storage engine are running.

        Refer to: https://docs.surrealdb.com/docs/integration/http#health

        :return: "OK" if everything is ok, or error text
        :raise SurrealConnectionError: if cant connect to http-endpoint
        """
        code, text = self.__get("health")
        return OK if code == HTTP_OK else text

    def version(self) -> str:
        """
        Get SurrealDB version

        Refer to: https://docs.surrealdb.com/docs/integration/http#version

        :return: version, for example, "surrealdb-1.1.1"
        :raise SurrealConnectionError: if cant connect to http-endpoint
        """
        return self.__get("version")[1]

    def __repr__(self) -> str:
        return f"Surreal(url={self._possible_url}, db_params={self.db_params}, timeout={self.timeout})"

    def __get(self, endpoint: str) -> Tuple[int, str]:
        try:
            with HttpClient(self._possible_url, timeout=DEFAULT_TIMEOUT).get(endpoint) as resp:
                status, text = resp.status, resp.read().decode(ENCODING)
                body = "is empty" if not text else text
                logger.info("Response from /%s, status_code: %s, body: %s", endpoint, status, body)
                if status != HTTP_OK:
                    logger.error("Status is %s for predicted http %s", status, self._possible_url)
                return status, text
        except HttpClientError:
            logger.error("Cant connect to %s", self._possible_url)
            raise SurrealConnectionError(f"Cant connect to {self._possible_url}{endpoint}\n"
                                         f"Is your SurrealDB started and work on {self._possible_url} ?\n"
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")



from surrealist.ql.statements.statement import FinishedStatement, Statement


class OnDuplicate(FinishedStatement):
    """
    Represents ON DUPLICATE part of the INSERT statement

    [ ON DUPLICATE KEY UPDATE @field = @value ... ]
    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/insert
    """
    def __init__(self, statement: Statement, action: str):
        super().__init__(statement)
        self._action = action

    def _clean_str(self):
        return f"{self._statement._clean_str()} ON DUPLICATE KEY UPDATE {self._action}"


class InsertUseDuplicate:
    def on_duplicate(self, action: str) -> OnDuplicate:
        """
        Add ON DUPLICATE clause to final statement

        ON DUPLICATE KEY UPDATE @field = @value ...

        :param action: raw representation of action
        :return: OnDuplicate object
        """
        return OnDuplicate(self, action)



from typing import List

from surrealist.connections import Connection
from surrealist.utils import OK
from .statement import Statement


class RebuildIndex(Statement):
    """
    Represents REBUILD statement, it should be able to use any statements from documentation

    Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/rebuild

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    REBUILD [INDEX [ IF EXISTS ] @name ON [ TABLE ] @table]
    """

    def __init__(self, connection: Connection, index_name: str, table_name: str, if_exists: bool = False):
        super().__init__(connection)
        self._name = index_name
        self._table_name = table_name
        self._if_exists = if_exists

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        exists = "" if not self._if_exists else " IF EXISTS"
        return f"REBUILD INDEX{exists} {self._name} ON TABLE {self._table_name}"



from typing import List, Optional, Union

from surrealist.connections.connection import Connection
from surrealist.enums import Algorithm
from surrealist.ql.statements import Create, Select
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class DefineAccessJwt(Statement):
    """
    A JWT access method allows accessing SurrealDB with a token signed by a trusted issuer.
    The contents of the token will be trusted by SurrealDB as long as it has been signed with a trusted credential.

    Refer to: https://surrealdb.com/docs/surrealql/statements/define/access/jwt

    DEFINE ACCESS [ OVERWRITE | IF NOT EXISTS ] @name
    ON [ NAMESPACE | DATABASE ]
    TYPE JWT [ ALGORITHM @algorithm KEY @key | URL @url ]
    [ AUTHENTICATE @expression ]
    [ DURATION FOR SESSION @duration ]
    """

    def __init__(self, connection: Connection, name: str):
        super().__init__(connection)
        self._name = name
        self._if_not_exists = None
        self._algo = None
        self._duration = None
        self._url = None
        self._auth = None

    def if_not_exists(self) -> "DefineAccessJwt":
        """
        Represents the IF NOT EXISTS statement.
        :return: DefineAccessJwt object
        """
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineAccessJwt":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def authenticate(self, raw_expression: str) -> "DefineAccessJwt":
        """
        Represents the AUTHENTICATE clause in a final statement. Expression will be inserted as is.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/access/jwt#with-authenticate-clause

        """
        self._auth = raw_expression
        return self

    def algorithm(self, algorithm: Algorithm, key: str) -> "DefineAccessJwt":
        """
        Represents the ALGORITHM clause in a final statement. If this method is called, the URL statement is ignored.

        :param algorithm: The algorithm to use, see :class:`surrealist.enums.Algorithm`
        :param key: The key to use
        :return: DefineAccessJwt object
        """
        self._algo = (algorithm, key)
        self._url = None
        return self

    def url(self, url: str) -> "DefineAccessJwt":
        """
        Represents the URL clause in a final statement. If this method is called, the ALGORITHM statement is ignored.
        """
        self._algo = None
        self._url = url
        return self

    def duration(self, duration: str) -> "DefineAccessJwt":
        """
        Represents the DURATION FOR SESSION clause in a final statement.
        Allowed values are: "1h", "1d", "1w", "1m", "1y"
        """
        self._duration = duration
        return self

    def validate(self) -> List[str]:
        if self._duration:
            if not any(self._duration.endswith(letter) for letter in ('h', 'd', 'w', 'm', 'y')):
                return [f"Wrong duration {self._duration}, allowed postfix are ('h', 'd', 'w', 'm', 'y')"]
            if ' ' in self._duration:
                return ["Wrong duration format, should be like 5m"]
        return [OK]

    def _clean_str(self):
        exists = ""
        if self._if_not_exists:
            exists = " IF NOT EXISTS"
        elif self._if_not_exists is False:
            exists = " OVERWRITE"
        access = ""
        if self._algo:
            access = f" ALGORITHM {self._algo[0].name} KEY '{self._algo[1]}'"
        if self._url:
            access = f" URL '{self._url}'"
        auth = "" if not self._auth else f" AUTHENTICATE {self._auth}"
        duration = f" DURATION FOR SESSION {self._duration}" if self._duration else ""
        return f"DEFINE ACCESS{exists} {self._name} ON DATABASE TYPE JWT{access}{auth}{duration}"


class DefineAccessRecord(Statement):
    """
    A record access method allows accessing SurrealDB as a record user.
    Record users allow SurrealDB to operate as a web database by offering mechanisms to define custom signin and signup
    logic as well as custom table and field permissions.

    Refer to: https://surrealdb.com/docs/surrealql/statements/define/access/record

        DEFINE ACCESS [ OVERWRITE | IF NOT EXISTS ] @name
      ON DATABASE TYPE RECORD
        [ SIGNUP @expression ]
        [ SIGNIN @expression ]
        [ AUTHENTICATE @expression ]
        [ WITH JWT
          [ ALGORITHM @algorithm KEY @key | URL @url ]
          [ WITH ISSUER KEY @key ]
        ]
      [ DURATION
        [ FOR TOKEN @duration ]
        [ FOR SESSION @duration ]
      ]
    """

    def __init__(self, connection: Connection, name: str):
        super().__init__(connection)
        self._name = name
        self._if_not_exists = None
        self._algo = None
        self._duration_token = None
        self._duration_session = None
        self._url = None
        self._signup = None
        self._signin = None
        self._auth = None

    def if_not_exists(self) -> "DefineAccessRecord":
        """
        Represents the IF NOT EXISTS statement.
        :return: DefineAccessJwt object
        """
        self._if_not_exists = True
        return self

    def overwrite(self) -> "DefineAccessRecord":
        """
        Adds OVERWRITE statement to the query
        :return: self
        """
        self._if_not_exists = False
        return self

    def signup(self, expression: Union[str, Create]) -> "DefineAccessRecord":
        """
        Represents the SIGNUP clause in a final statement. Can be raw string or Create object
        """
        self._signup = expression
        return self

    def signin(self, expression: Union[str, Select]) -> "DefineAccessRecord":
        """
        Represents the SIGNUP clause in a final statement. Can be raw string or Select object
        """
        self._signin = expression
        return self

    def authenticate(self, raw_expression: str) -> "DefineAccessRecord":
        """
        Represents the AUTHENTICATE clause in a final statement. Expression will be inserted as is.

        Refer to:
        https://surrealdb.com/docs/surrealql/statements/define/access/record#with-authenticate-clause


        """
        self._auth = raw_expression
        return self

    def algorithm(self, algorithm: Algorithm, key: str, issuer_key: Optional[str] = None) -> "DefineAccessRecord":
        """
        Represents the ALGORITHM clause in a final statement. If this method is called, the URL statement is ignored.

        :param algorithm: The algorithm to use, see :class:`surrealist.enums.Algorithm`
        :param key: The key to use
        :param issuer_key: The issuer key to use
        :return: DefineAccessRecord object
        """
        self._algo = (algorithm, key, issuer_key)
        self._url = None
        return self

    def url(self, url: str, issuer_key: Optional[str] = None) -> "DefineAccessRecord":
        """
        Represents the URL clause in a final statement. If this method is called, the ALGORITHM statement is ignored.
        :param url: The URL to use
        :param issuer_key: The issuer key to use
        :return: DefineAccessRecord object
        """
        self._algo = None
        self._url = (url, issuer_key)
        return self

    def duration_for_token(self, duration: str) -> "DefineAccessRecord":
        """
        Represents the DURATION FOR TOKEN clause in a final statement.
        Allowed values are: "1h", "1d", "1w", "1m", "1y"
        """
        self._duration_token = duration
        return self

    def duration_for_session(self, duration: str) -> "DefineAccessRecord":
        """
        Represents the DURATION FOR SESSION clause in a final statement.
        Allowed values are: "1h", "1d", "1w", "1m", "1y"
        """
        self._duration_session = duration
        return self

    def validate(self) -> List[str]:
        for duration in (self._duration_token, self._duration_session):
            if duration:
                if not any(duration.endswith(letter) for letter in ('h', 'd', 'w', 'm', 'y')):
                    return [f"Wrong duration {duration}, allowed postfix are ('h', 'd', 'w', 'm', 'y')"]
                if ' ' in duration:
                    return ["Wrong duration format, should be like 5m"]
        return [OK]

    def _clean_str(self):
        exists = ""
        if self._if_not_exists:
            exists = " IF NOT EXISTS"
        elif self._if_not_exists is False:
            exists = " OVERWRITE"
        sin = ""
        if self._signin:
            sin = f" SIGNIN {self._signin}" if isinstance(self._signin, str) else f" SIGNIN {self._signin._clean_str()}"
        sup = ""
        if self._signup:
            sup = f" SIGNUP {self._signup}" if isinstance(self._signup, str) else f" SIGNUP {self._signup._clean_str()}"
        access = ""
        if self._algo:
            access = f" ALGORITHM {self._algo[0].name} KEY '{self._algo[1]}'"
            if self._algo[2]:
                access = f"{access} WITH ISSUER KEY '{self._algo[2]}'"
        duration_t = f"DURATION FOR TOKEN {self._duration_token}" if self._duration_token else ""
        duration_s = f"DURATION FOR SESSION {self._duration_session}" if self._duration_session else ""
        duration = f' {", ".join((duration_t, duration_s))}'
        if not duration_t and not duration_s:
            duration = ""
        if self._url:
            access = f" URL '{self._url[0]}'"
            if self._url[1]:
                access = f"{access} WITH ISSUER KEY '{self._url[1]}'"
        auth = "" if not self._auth else f" AUTHENTICATE {self._auth}"
        return f"DEFINE ACCESS{exists} {self._name} ON DATABASE TYPE RECORD{sin}{sup}{auth}{access}{duration}"



import json
from typing import List

from surrealist.ql.statements.statement import FinishedStatement, Statement
from surrealist.utils import OK


class Comment(FinishedStatement):
    """
    Represents a COMMENT clause in statements.
    """

    def __init__(self, statement: Statement, text: str):
        super().__init__(statement)
        self._text = text

    def validate(self) -> List[str]:
        if not self._text:
            return ["Comment text is empty"]
        return [OK]

    def _clean_str(self):
        text = f" COMMENT {json.dumps(self._text)}" if self._text else ""
        return f"{self._statement._clean_str()}{text}"


class CanUseComment:
    def comment(self, comment: str) -> Comment:
        """
        Adds COMMENT statement to the query
        :param comment: comment text
        :return: self
        """
        return Comment(self, comment)


class Parallel(FinishedStatement):
    """
    Represents a 'PARALLEL' statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} PARALLEL"


class CanUseParallel:
    def parallel(self) -> Parallel:
        return Parallel(self)


class Timeout(FinishedStatement, CanUseParallel):
    """
    Represents a 'TIMEOUT' statement. Allowed durations are (ms, s, m)
    """

    def __init__(self, statement: Statement, duration: str):
        super().__init__(statement)
        self._duration = duration

    def _validate(self):
        if not any(self._duration.endswith(letter) for letter in ('ms', 's', 'm')):
            return f"Wrong duration {self._duration}, allowed postfix are (ms, s, m)"
        if ' ' in self._duration:
            return "Wrong duration format, should be like 5s"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} TIMEOUT {self._duration}"


class CanUseTimeout(CanUseParallel):
    def timeout(self, duration: str) -> Timeout:
        return Timeout(self, duration)


class ReturnNone(FinishedStatement, CanUseTimeout):
    """
    Represents a 'RETURN NONE' statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN NONE"


class ReturnBefore(FinishedStatement, CanUseTimeout):
    """
    Represents a 'RETURN BEFORE' statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN BEFORE"


class ReturnAfter(FinishedStatement, CanUseTimeout):
    """
    Represents a 'RETURN AFTER' statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN AFTER"


class ReturnDiff(FinishedStatement, CanUseTimeout):
    """
    Represents a 'RETURN DIFF' statement.
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN DIFF"


class Return(FinishedStatement, CanUseTimeout):
    """
    Represents a 'RETURN' statement.
    """

    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement._clean_str()
        args = ", ".join(self._args)
        return f"{self._statement._clean_str()} RETURN {args}"


class CanUseReturn(CanUseTimeout):
    def return_diff(self) -> ReturnDiff:
        return ReturnDiff(self)

    def return_none(self) -> ReturnNone:
        return ReturnNone(self)

    def return_before(self) -> ReturnBefore:
        return ReturnBefore(self)

    def return_after(self) -> ReturnAfter:
        return ReturnAfter(self)

    def returns(self, *args: str) -> Return:
        return Return(self, *args)


class Where(FinishedStatement, CanUseReturn):
    """
    Represents a 'WHERE' statement.
    """

    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._predicate = predicate

    def _clean_str(self):
        return f"{self._statement._clean_str()} WHERE {self._predicate}"


class CanUseWhere(CanUseReturn):
    def where(self, predicate: str) -> Where:
        return Where(self, predicate)



from logging import getLogger
from pathlib import Path
from typing import Tuple, Dict, Optional, Union, Any, BinaryIO

from surrealist.clients.http_client import HttpClient
from surrealist.connections.connection import Connection, connected
from surrealist.enums import Transport
from surrealist.errors import (CompatibilityError, HttpConnectionError, HttpClientError, SurrealConnectionError)
from surrealist.result import SurrealResult, to_result
from surrealist.utils import (ENCODING, DEFAULT_TIMEOUT, HTTP_OK, NS, DB, AC)

logger = getLogger("surrealist.connections.http")


class HttpConnection(Connection):
    """
    Represents http transport and abilities to work with SurrealDb. It is not a recommended connection, use websocket in 
    any doubt; this connection exists for compatibility reasons. Some features as live query are not possible on this
    transport, but import and export are.

    Refer to surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#transports
    Refer to: https://docs.surrealdb.com/docs/integration/http

    Each method call creates a new http short-live connection.

    On creating, this object tries to create a connection with specified data and will raise exception on fail.
    """

    def __init__(self, url: str, db_params: Optional[Dict] = None, credentials: Tuple[str, str] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        super().__init__(db_params, credentials, timeout)
        self._url = url
        self._http_client = HttpClient(url, headers=db_params, credentials=credentials, timeout=timeout)
        self._sign(credentials, db_params, url)
        self._connected = True
        masked_creds = None if not credentials else (credentials[0], "******")
        logger.info("Connected to %s, params: %s, credentials: %s, timeout: %s", url, db_params, masked_creds, timeout)

    def _use_rpc(self, data) -> SurrealResult:
        _, text = self._rpc(data)
        return to_result(text)

    def _sign(self, credentials, db_params, url):
        user, password, ns, db, ac = None, None, None, None, None
        if credentials:
            user, password = credentials
        if db_params:
            ns = db_params.get(NS)
            db = db_params.get(DB)
            ac = db_params.get(AC)
        try:
            result = self._signin(user, password, namespace=ns, database=db, access=ac)
            if result.is_error():
                logger.error("Cant sign in to %s with given credentials", url)
                raise SurrealConnectionError(f"Cant sign in to {url} with given credentials\n"
                                             f"Info: {result.additional_info}\n")
            self._http_client.set_token(result.result)
            self._token = result.result
        except HttpClientError:
            logger.error("Cant connect to %s", url)
            raise SurrealConnectionError(f"Cant connect to {url}\n"
                                         f"Is your SurrealDB started and work on that url? "
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")

    def transport(self) -> Transport:
        """
        Returns the transport type for http connection
        """
        return Transport.HTTP

    @connected
    def import_data(self, path: Union[str, Path]) -> SurrealResult:
        """
        This method imports a SurrealQL script file into a local or remote SurrealDB database server.

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#import
        https://docs.surrealdb.com/docs/cli/import

        Refer to: https://docs.surrealdb.com/docs/integration/http#import
        :param path: path to file for import 
        :return: result of request
        """
        with open(path, 'rb') as file:
            logger.info("Operation: IMPORT. Path: %s", path)
            _, text = self._simple_request("POST", "import", file, type_of_content="FILE")
        return to_result(text)

    @connected
    def export(self) -> str:
        """
        This method exports all data for a specific namespace and database.

        Refer to: https://docs.surrealdb.com/docs/integration/http#export

        Refer to: https://docs.surrealdb.com/docs/cli/export/


        :return: text of the exported file to use or save
        """
        text = raise_if_not_http_ok(self._simple_get("export"))
        return text

    @connected
    def ml_import(self, path: Union[str, Path]) -> SurrealResult:
        """
        This method imports a SurrealQL ML file into a local or remote SurrealDB database server.

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#ml-import
        https://docs.surrealdb.com/docs/cli/ml/import

        :param path: path to file for import
        :return: result of request
        """
        with open(path, 'rb') as file:
            logger.info("Operation: ML IMPORT. Path: %s", path)
            _, text = self._simple_request("POST", "ml/import", file.read().decode(ENCODING), type_of_content="STR")
        return to_result(text)

    @connected
    def ml_export(self, name: str, version: str) -> str:
        """
        This method exports a SurrealML machine learning model from a specific namespace and database. As machine
        learning files can be large, the endpoint outputs a chunked HTTP response.

        Refer to: https://docs.surrealdb.com/docs/integration/http#ml-export

        Refer to: https://docs.surrealdb.com/docs/cli/ml/export

        Example:
        http_connection.ml_export("prediction", "1.0.0")

        :param name: name of ML model
        :param version: version of ML model
        :return:  text of the exported file to use or save
        """
        logger.info("Operation: ML EXPORT. Name %s Version %s", name, version)
        text = raise_if_not_http_ok(self._simple_get(f"ml/export/{name}/{version}"))
        return text

    @connected
    def use(self, namespace: str, database: Optional[str] = None) -> None:
        """
        This method specifies the namespace and optional database for the current connection. For http-connection, it
        actually means using headers "surreal-ns" and "surreal-db", no checks for namespace or database to be existent
        will be made

        Example:
        http_connection.use("test", "test") # use test namespace and test database for this connection

        :param namespace: name of the namespace to use
        :param database: name of the database to use (optional)
        :return: None
        """
        logger.info("Operation: USE. Namespace: %s, database %s", namespace, database or "None")
        self._db_params = {NS: namespace}
        if database:
            self._db_params[DB] = database
        self._http_client.set_db_params(self._db_params)

    def live(self, table_name, callback, return_diff: bool = False):
        """
        Http transport cannot use live queries, you should use websocket transport for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport cannot use live queries, use websockets instead"
        logger.error(message)
        raise CompatibilityError(message)

    def custom_live(self, custom_query, callback):
        """
        Http transport cannot use live queries, you should use websocket transport for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport cannot use custom live queries, use websockets instead"
        logger.error(message)
        raise CompatibilityError(message)

    def kill(self, live_query_id: str):
        """
        Http transport cannot use KILL operation, you should use websocket transport for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport cannot use kill, use websockets instead"
        logger.error(message)
        raise CompatibilityError(message)

    def let(self, name: str, value: Any):
        """
        Http transport cannot use LET method, you should use websocket for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport cannot use LET method, you should use websocket for that"
        logger.error(message)
        raise CompatibilityError(message)

    def unset(self, name: str):
        """
        Http transport cannot use UNSET method, you should use websocket for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport cannot use UNSET method, you should use websocket for that"
        logger.error(message)
        raise CompatibilityError(message)

    def _simple_get(self, endpoint: str) -> Tuple[int, str]:
        with self._http_client.get(endpoint) as resp:
            status, text = resp.status, resp.read().decode(ENCODING)
            _body = "is empty" if not text else text
            logger.info("Response from /%s, status_code: %s, body: %s", endpoint, status, _body)
            return status, text

    def _rpc(self, data: Union[Dict, str]) -> Tuple[int, str]:
        return self._simple_request("POST", "rpc", data)

    def _simple_request(self, method, endpoint: str, data: Union[Dict, str, BinaryIO],
                        type_of_content: str = "JSON") -> Tuple[int, str]:
        with self._http_client.request(method, data, endpoint, type_of_content=type_of_content) as resp:
            status, text = resp.status, resp.read().decode(ENCODING)
            if type_of_content == "FILE":
                data.close()
            _body = "is empty" if not text else text
            logger.info("Response from /%s, status_code: %s, body %s", endpoint, status, _body)
        return status, text


def raise_if_not_http_ok(result: Tuple[int, str]) -> str:
    """
    Helper for methods which need only success responses

    :param result: pair of status_code and text from the http response
    :return: text of the response if status is 200, else raise
    :raise HttpConnectionError: if status code is not 200
    """
    status, text = result
    if status != HTTP_OK:
        logger.error("Status code is %s, check your data! Info %s", status, text)
        text = text.replace(',', ',\n')
        raise HttpConnectionError(f"Status code is {status}, check your data! Info:\n{text}")
    return text



from typing import Optional, Tuple, Union, List

from surrealist.connections import Connection
from surrealist.utils import OK, StrOrRecord, get_table_or_record_id
from .select_statements import SelectUseIndex, SelectUseTempfiles
from .statement import Statement, IterableStatement


class Select(IterableStatement, SelectUseIndex, SelectUseTempfiles):
    """
    Represents SELECT statement, it should be able to use any statements from documentation.
    It can use iterator to traverse results

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/select

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_select_examples.py

    Iterator example: https://github.com/kotolex/surrealist/tree/master/examples/surreal_ql/iterator.py

    SELECT [ VALUE ] @fields [ AS @alias ]
    [ OMIT @fields ...]
    FROM [ ONLY ] @targets
    [ WITH [ NOINDEX | INDEX @indexes ... ]]
    [ WHERE @conditions ]
    [ SPLIT [ AT ] @field ... ]
    [ GROUP [ BY ] @fields ... ]
    [ ORDER [ BY ]
        @fields [
            RAND()
            | COLLATE
            | NUMERIC
        ] [ ASC | DESC ] ...
    ]
    [ LIMIT [ BY ] @limit ]
    [ START [ AT ] @start ]
    [ FETCH @fields ... ]
    [ TIMEOUT @duration ]
    [ PARALLEL ]
    [TEMPFILES]
    [ EXPLAIN [ FULL ]];
    """

    def __init__(self, connection: Connection, table_name: Union[str, Statement], *args,
                 alias: Optional[List[Tuple[str, Union[str, Statement]]]] = None,
                 value: Optional[str] = None):
        self._connection = connection
        super().__init__(self)
        self._table_name = table_name if isinstance(table_name, str) else f"({table_name._clean_str()})"
        self._args = args
        self._what = ""
        self._from = self._table_name
        self._only = False
        self._value = value
        self._as = alias
        self._update_args()
        self._omit = None

    def _update_args(self):
        if self._args:
            self._what = ", ".join(self._args)
        if self._as:
            what = "" if not self._what else f"{self._what}, "
            result = []
            for alias, value in self._as:
                value = value if not isinstance(value, Statement) else f"({value._clean_str()})"
                result.append(f"{value} AS {alias}")
            final_str = ", ".join(result)
            self._what = f"{what}{final_str}"
        if self._value:
            self._what = f"VALUE {self._value}"

    def by_id(self, record_id: StrOrRecord) -> "Select":
        """
        Will select by given id SELECT @fields from table:record_id

        :param record_id: id of the record to select
        """
        self._from = get_table_or_record_id(self._table_name, record_id)
        return self

    def only(self) -> "Select":
        """
        Include ONLY statement for the query
        """
        self._only = True
        return self

    def omit(self, *fields: Tuple[str]) -> "Select":
        """
        Include OMIT statement for the query
        :param fields: fields to omit
        """
        self._omit = fields
        return self

    def validate(self) -> List[str]:
        if self._only and ":" not in self._from:
            return ["Expected a single result output when using the ONLY keyword"]
        return [OK]

    def _clean_str(self):
        only = "ONLY " if self._only else ""
        what = "*" if not self._what else self._what
        omit = ""
        if self._omit:
            omitted = ", ".join(self._omit)
            omit = f" OMIT {omitted}"
        return f"SELECT {what}{omit} FROM {only}{self._from}"

    def __repr__(self):
        return f"Select({self._table_name=}, {self._args=}, {self._as=}, {self._value=})"



import datetime
import json
import re
import uuid
from typing import Union, Dict, Optional, List, Tuple, Any

from .errors import SurrealRecordIdError
from .record_id import RecordId

StrOrRecord = Union[str, RecordId]  # Represents a string or a RecordId for queries
StrOrInt = Union[str, int]  # Represents datetime or versionstamp for SHOW statement

ENCODING = "UTF-8"
OK = "OK"
ERR = "ERR"
DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DATE_FORMAT_NS = "%Y-%m-%dT%H:%M:%S.%fZ"
HTTP_OK = 200  # status code for success
DEFAULT_TIMEOUT = 15  # timeout in seconds for basic operations
LOG_FORMAT = '%(asctime)s : %(threadName)s : %(name)s : %(levelname)s : %(message)s'  # use it for logs
NS = "NS"
DB = "DB"
AC = "AC"


def get_uuid() -> str:
    """
    Helper to generate uuid for records or testing purposes

    :return: uuid string representation
    """
    return str(uuid.uuid4())


def mask_pass(text: str) -> str:
    """
    Mask the passwords in logs. Replace all 'pass':'any_pass' in logs to 'pass':'******'. Works both with ' and "
    :param text: text before putting it to log
    :return: text without visible passwords
    """
    return re.sub(r"(?ms)(?<=['\"]pass['\"]: ['\"]).*?(?=['\"])", '******', text)


def to_surreal_datetime_str(dt: datetime.datetime) -> str:
    """
    Convert datetime to string in Surreal format, for example: d'2024-04-18T11:34:41.665249Z'
    :param dt: datetime object
    :return: string representation of the datetime
    """
    return f"d'{dt.strftime(DATE_FORMAT_NS)}'"


def to_datetime(dt_str: str) -> datetime.datetime:
    """
    Convert string in Surreal format to datetime object
    :param dt_str: string datetime representation in Surreal format (e.g. d'2024-04-18T11:34:41.665249Z')
    :return: datetime object
    """
    if "d'" in dt_str:
        dt_str = dt_str.rstrip("'").replace("d'", "")
    elif 'd"' in dt_str:
        dt_str = dt_str.rstrip('"').replace('d"', '')
    return datetime.datetime.strptime(dt_str, DATE_FORMAT_NS)


def clean_dates(data: str) -> str:
    """
    Get surreal dates with d' prefix out of quotes
    :param data: some Surreal query
    :return: data with valid Surreal dates
    """
    return re.sub(r'["\'](d(["\']).+?)["\']+', r'\1\2', data)


def safe_dumps(data: Any) -> str:
    """
    Convert data to json string with special logic for RecordId (if it exists)
    We have to do it because since version 2.0 of SurrealDB it never converts string to record_id
    :param data: data to convert, dict, list, tuple or JSON serializable object expected here
    :return: string
    """
    if isinstance(data, RecordId):
        return data.to_valid_string()
    if isinstance(data, List):
        return list_to_json_str(data)
    if isinstance(data, Tuple):
        return tuple_to_json_str(data)
    if isinstance(data, Dict):
        return dict_to_json_str(data)
    return json.dumps(data)


def dict_to_json_str(data: Dict) -> str:
    """
    Convert dict to json string with special logic for RecordId (if it exists)
    We have to do it because since version 2.0 of SurrealDB it never converts string to record_id
    :param data: dict to convert
    :return: string
    """
    ids = {k: v for k, v in data.items() if isinstance(v, (RecordId, List, Dict))}
    if not ids:
        return json.dumps(data)
    without_ids = {k: v for k, v in data.items() if k not in ids}
    first = json.dumps(without_ids) if without_ids else ""
    join = ", ".join([f'"{k}": {safe_dumps(v)}' for k, v in ids.items()])
    if first:
        first = first.rstrip("}")
        first = f"{first}, "
    else:
        first = "{"
    return f"{first}{join}}}"


def list_to_json_str(data: List) -> str:
    """
    Convert a list to json string with special logic for RecordId (if it exists)
    We have to do it because since version 2.0 of SurrealDB it never converts string to record_id
    :param data: list of pairs to convert
    :return: string
    """
    ids = [e for e in data if isinstance(e, (RecordId, List, Dict, Tuple))]
    if not ids:
        return json.dumps(data)
    without_ids = [e for e in data if e not in ids]
    first = json.dumps(without_ids) if without_ids else ""
    join = ", ".join(safe_dumps(e) for e in ids)
    if first:
        first = first.rstrip("]")
        first = f"{first}, "
    else:
        first = "["
    return f"{first}{join}]"


def tuple_to_json_str(values: Tuple) -> str:
    """
    Convert a tuple to json string with special logic for RecordId (if it exists)
    We have to do it because since version 2.0 of SurrealDB it never converts string to record_id
    :param values: tuple to convert
    :return: string
    """
    ids = [e for e in values if isinstance(e, (RecordId, List, Dict, Tuple))]
    if not ids:
        return f'({json.dumps(values).lstrip("[").rstrip("]")})'
    without_ids = [e for e in values if e not in ids]
    first = json.dumps(without_ids)
    first = first.lstrip("[").rstrip("]")
    if first:
        first = f"{first}, "
    join = ", ".join(f'{safe_dumps(e)}' for e in ids)
    return f"({first}{join})"


def get_table_or_record_id(table_name: str, record_id: Optional[StrOrRecord]) -> str:
    """
    A helper for backward compatibility and converting string to record_id
    :param table_name: name of table
    :param record_id: string or record_id
    :return: valid record_id representation
    """
    if record_id is not None:
        if isinstance(record_id, (str, int)):
            record_id = RecordId(str(record_id), table=table_name)
        part = record_id.table_part
        if table_name != part:
            raise SurrealRecordIdError(f"Table name is different from id, we expect {table_name}, but got {part}")
        return record_id.to_valid_string()
    return table_name



from typing import Optional, Dict

from surrealist.utils import safe_dumps


def combine(result: Optional[str], kwargs: Dict) -> str:
    """
    Combine optional raw string representation with optional keyword-arguments. Used in relate/create/update statements
    for set methods

    :param result: raw string representation for query
    :param kwargs: optional parameters
    :return: string combined
    """
    first = result if result else ''
    second = ", ".join(f"{k} = {safe_dumps(v)}" for k, v in kwargs.items()) if kwargs else ''
    args = ', '.join(element for element in (first, second) if element)
    return args



from .update import Update


class Upsert(Update):
    """
    Represents an UPSERT statement.

    The UPSERT statement can be used to modify records in the database if they already exist. If the record does not
    exist, it will be created.
    This is different from the UPDATE statement, which will fail if the record does not exist.

    Refer to: https://surrealdb.com/docs/surrealql/statements/upsert

    UPSERT [ ONLY ] @targets
    [ CONTENT @value
      | MERGE @value
      | PATCH @value
      | SET @field = @value ...
    ]
    [ WHERE @condition ]
    [ RETURN NONE | RETURN BEFORE | RETURN AFTER | RETURN DIFF | RETURN @statement_param, ... ]
    [ TIMEOUT @duration ]
    [ PARALLEL ];
    """

    def _clean_str(self):
        return f"UPSERT{super()._clean_str().lstrip('UPDATE')}"
